import pytest

from tickit.utils.topic_naming import input_topic, output_topic


def test_input_affixes():
    assert "tickit-my_device-in" == input_topic("my_device")


def test_input_raises_empty():
    with pytest.raises(ValueError):
        input_topic("")


def test_output_affixes():
    assert "tickit-my_device-out" == output_topic("my_device")


def test_output_raises_empty():
    with pytest.raises(ValueError):
        output_topic("")



from dataclasses import dataclass

from apischema import deserialize
from apischema.json_schema import deserialization_schema

from tickit.utils.configuration.configurable import as_tagged_union


@as_tagged_union
class MyBase:
    pass


@dataclass
class MyClass(MyBase):
    a: int
    b: str


@dataclass
class MyOtherClass(MyBase):
    a: int
    c: float


def test_tagged_union_deserializes():
    expected = MyClass(a=1, b="foo")
    actual = deserialize(MyBase, {"test_configurable.MyClass": {"a": 1, "b": "foo"}})
    assert actual == expected


def test_deserialization_schema():
    assert deserialization_schema(MyBase) == {
        "$schema": "http://json-schema.org/draft/2020-12/schema#",
        "additionalProperties": False,
        "maxProperties": 1,
        "minProperties": 1,
        "properties": {
            "test_configurable.MyClass": {
                "additionalProperties": False,
                "properties": {"a": {"type": "integer"}, "b": {"type": "string"}},
                "required": ["a", "b"],
                "type": "object",
            },
            "test_configurable.MyOtherClass": {
                "additionalProperties": False,
                "properties": {"a": {"type": "integer"}, "c": {"type": "number"}},
                "required": ["a", "c"],
                "type": "object",
            },
        },
        "type": "object",
    }



from collections import defaultdict
from typing import (
    Any,
    Awaitable,
    Callable,
    DefaultDict,
    Generic,
    Iterable,
    List,
    NamedTuple,
    NewType,
    Set,
    TypeVar,
)

from tickit.core.state_interfaces import state_interface
from tickit.utils.singleton import Singleton

#: A consumable value
C = TypeVar("C")
#: A producable value
P = TypeVar("P")


class Message(NamedTuple):
    """An immutable data container for internal messages."""

    value: Any


#: A list of messages
Messages = NewType("Messages", List[Message])


class InternalStateServer(metaclass=Singleton):
    """A singleton, in memory, publish/subscribe message store and router.

    A singleton, in memory, publish/subscribe message store and router. The single
    instance of this class holds a mapping of topics and a list of messages which have
    been pushed to them and a mapping of consumers which subscribe to the topics. Upon
    subscribing to a topic, a consumer is sent all of the messages which were
    previously registered against the topic. Upon pushing to a topic, the message is
    added to the store and immediately forwarded to each of the consumers which
    subscribe to the topic.
    """

    _topics: DefaultDict[str, Messages] = defaultdict(lambda: Messages(list()))
    _subscribers: DefaultDict[str, Set["InternalStateConsumer"]] = defaultdict(set)

    async def push(self, topic: str, message: Message) -> None:
        """Asynchronous method which propagates a message to subscribers and stores it.

        An asynchronous method which propagates the given message to consumers which
        subscribe to the topic and stores the message in the topic -> messages mapping.

        Args:
            topic (str): The topic on which the the message should be propagated and
                stored.
            message (Message): The message which should be propagated and stored on the
                topic.
        """
        self._topics[topic].append(message)
        for subscriber in self._subscribers[topic]:
            await subscriber.add_message(message)

    async def subscribe(self, consumer: "InternalStateConsumer", topics: Iterable[str]):
        """Subscribes the consumer to the given topics, so it is notified a message is
        added.

        An asynchronous method which adds a consumer to the subscriber list of each
        topic in the topics iterable. On subscription, previous messages on the topic
        are immediately passed to the consumer.

        Args:
            consumer (InternalStateConsumer): The consumer which is subscribing to the
                topic.
            topics (Iterable[str]): The topic which the consumer is to be subscribed to.
        """
        for topic in topics:
            self._subscribers[topic].add(consumer)
            for message in self._topics[topic]:
                await consumer.add_message(message)

    def create_topic(self, topic: str) -> None:
        """Creates a new topic as an empty message list.

        Args:
            topic (str): The topic to be created.
        """
        assert topic not in self._topics.keys()
        self._topics[topic] = Messages(list())

    def remove_topic(self, topic: str) -> None:
        """Removes an existing topic and the corresponding message list.

        Args:
            topic (str): The topic to be removed.
        """
        assert topic in self._topics.keys()
        del self._topics[topic]

    @property
    def topics(self) -> List[str]:
        """A list of topics which currently exist.

        Returns:
            List[str]: A list of topics which currently exist.
        """
        return list(self._topics.keys())


@state_interface.add("internal", False)
class InternalStateConsumer(Generic[C]):
    """An internal, singleton based, implementation of the StateConsumer protocol.

    A internal, singleton based, implementation of the StateConsumer protocol, this
    consumer can subscribe to InternalStateServer topics, upon recieving a message the
    consumer passes the value to the callback function passed during initialization, if
    a topic is subscribed to which does not yet exist it is created.
    """

    def __init__(self, callback: Callable[[C], Awaitable[None]]) -> None:
        """Gets an instance of the InternalStateServer for use in subscribe.

        Args:
            callback (Callable[[C], Awaitable[None]]): An asynchronous handler function
                for consumed values.
        """
        self.server = InternalStateServer()
        self.callback = callback

    async def subscribe(self, topics: Iterable[str]) -> None:
        """Subscribes the consumer to the given topics, new messages are passed to the
        callback.

        Args:
            topics (Iterable[str]): An iterable of topics to subscribe to.
        """
        await self.server.subscribe(self, topics)

    async def add_message(self, message: Message) -> None:
        """Adds a message to the consumer, triggering a callback.

        Args:
            message (Message): The message to be added to the consumer.
        """
        await self.callback(message.value)


@state_interface.add("internal", False)
class InternalStateProducer(Generic[P]):
    """An internal, singleton based, implementation of the StateProducer protocol.

    An internal, singleton based, implementation of the StateProducer protocol, this
    producer can produce a value to a InternalStateServer topic, if the topic does not
    yet exist it is created.
    """

    def __init__(self) -> None:
        """Gets an instance of the InternalStateServer for use in produce."""
        self.server = InternalStateServer()

    async def produce(self, topic: str, value: P) -> None:
        """Produces a value to the provided topic.

        Args:
            topic (str): The topic to which the value should be sent.
            value (P): The value to send to the provided topic.
        """
        await self.server.push(topic, Message(value=value))



import pytest

from tickit.devices.eiger.monitor.eiger_monitor import EigerMonitor


@pytest.fixture
def filewriter() -> EigerMonitor:
    return EigerMonitor()


def test_eiger_monitor_constructor():
    EigerMonitor()



import json

import aiohttp
import pytest
from aiohttp import web
from mock import MagicMock, Mock
from mock.mock import create_autospec

from tickit.devices.eiger.eiger import EigerDevice
from tickit.devices.eiger.eiger_adapters import EigerRESTAdapter
from tickit.devices.eiger.eiger_settings import EigerSettings
from tickit.devices.eiger.eiger_status import EigerStatus, State


@pytest.fixture
def mock_status() -> MagicMock:
    status = create_autospec(EigerStatus, instance=True)
    status.state = State.NA
    return status


@pytest.fixture
def mock_settings() -> MagicMock:
    settings = create_autospec(EigerSettings, instance=True)
    settings.count_time = {
        "value": 0.1,
        "metadata": {"value_type": Mock(value="int"), "access_mode": Mock(value="rw")},
    }
    return settings


@pytest.fixture
def mock_eiger(mock_status: MagicMock, mock_settings: MagicMock) -> MagicMock:
    mock_eiger = create_autospec(EigerDevice, instance=True)
    mock_eiger.status = mock_status
    mock_eiger.settings = mock_settings
    return mock_eiger


@pytest.fixture
def raise_interrupt():
    async def raise_interrupt():
        return False

    return Mock(raise_interrupt)


@pytest.fixture
def eiger_adapter(mock_eiger: MagicMock) -> EigerRESTAdapter:
    return EigerRESTAdapter("0.0.0.0", 8081)


def test_eiger_adapter_contructor():
    EigerRESTAdapter(mock_eiger, raise_interrupt)


@pytest.fixture()
def mock_request():
    mock_request = MagicMock(web.Request)
    return mock_request


@pytest.mark.asyncio
@pytest.mark.parametrize("tickit_task", ["examples/configs/eiger.yaml"], indirect=True)
async def test_eiger_system(tickit_task):
    commands = {
        "initialize": {"sequence_id": 1},
        "disarm": {"sequence_id": 3},
        "cancel": {"sequence_id": 5},
        "abort": {"sequence_id": 6},
    }

    url = "http://0.0.0.0:8081/detector/api/1.8.0/"
    headers = {"content-type": "application/json"}

    filewriter_url = "http://0.0.0.0:8081/filewriter/api/1.8.0/"
    monitor_url = "http://0.0.0.0:8081/monitor/api/1.8.0/"
    stream_url = "http://0.0.0.0:8081/stream/api/1.8.0/"

    async def get_status(status, expected):
        async with session.get(url + f"status/{status}") as resp:
            assert expected == json.loads(str(await resp.text()))["value"]

    async with aiohttp.ClientSession() as session:
        await get_status(status="state", expected="na")

        # Test setting config var before Eiger set up
        data = '{"value": "test"}'
        async with session.put(
            url + "config/element", headers=headers, json=data
        ) as resp:
            assert json.loads(str(await resp.text())) == {"sequence_id": 7}

        # Test each command
        for key, value in commands.items():
            async with session.put(url + f"command/{key}") as resp:
                assert value == json.loads(str(await resp.text()))

        await get_status(status="doesnt_exist", expected="None")

        await get_status(status="board_000/th0_temp", expected=24.5)

        await get_status(status="board_000/doesnt_exist", expected="None")

        await get_status(status="builder/dcu_buffer_free", expected=0.5)

        await get_status(status="builder/doesnt_exist", expected="None")

        # Test Eiger in IDLE state
        await get_status(status="state", expected="idle")

        async with session.get(url + "config/doesnt_exist") as resp:
            assert json.loads(str(await resp.text()))["value"] == "None"

        data = '{"value": "test"}'
        async with session.put(
            url + "config/doesnt_exist", headers=headers, json=data
        ) as resp:
            assert json.loads(str(await resp.text())) == {"sequence_id": 9}

        async with session.get(url + "config/element") as resp:
            assert json.loads(str(await resp.text()))["value"] == "Co"

        data = '{"value": "Li"}'
        async with session.put(
            url + "config/element", headers=headers, json=data
        ) as resp:
            assert json.loads(str(await resp.text())) == {"sequence_id": 8}

        async with session.get(url + "config/photon_energy") as resp:
            assert json.loads(str(await resp.text()))["value"] == 54.3

        async with session.get(filewriter_url + "config/mode") as resp:
            assert "enabled" == json.loads(str(await resp.text()))["value"]

        async with session.get(filewriter_url + "status/state") as resp:
            assert "ready" == json.loads(str(await resp.text()))["value"]

        async with session.get(monitor_url + "config/mode") as resp:
            assert "enabled" == json.loads(str(await resp.text()))["value"]

        async with session.get(monitor_url + "status/error") as resp:
            assert [] == json.loads(str(await resp.text()))["value"]

        async with session.get(stream_url + "config/mode") as resp:
            assert "enabled" == json.loads(str(await resp.text()))["value"]

        async with session.get(stream_url + "status/state") as resp:
            assert "ready" == json.loads(str(await resp.text()))["value"]

        async with session.put(url + "command/arm") as resp:
            assert {"sequence_id": 2} == json.loads(str(await resp.text()))

        async with session.put(url + "command/trigger") as resp:
            assert {"sequence_id": 4} == json.loads(str(await resp.text()))



from dataclasses import dataclass
from typing import AsyncIterable, Generic, Optional, TypeVar

from tickit.core.adapter import Adapter, Interpreter, RaiseInterrupt, Server
from tickit.core.device import Device

#: Message type
T = TypeVar("T")


@dataclass
class ComposedAdapter(Adapter, Generic[T]):
    """An adapter implementation which delegates to a server and interpreter.

    An adapter implementation which delegates the hosting of an external messaging
    protocol to a server and message handling to an interpreter.
    """

    server: Server
    interpreter: Interpreter

    async def on_connect(self) -> AsyncIterable[Optional[T]]:
        """An overridable asynchronous iterable which yields messages on client
        connection.

        Returns:
            AsyncIterable[Optional[T]]: An asynchronous iterable of messages.
        """
        if False:
            yield None

    async def handle_message(self, message: T) -> AsyncIterable[Optional[T]]:
        """Delegates message handling to the interpreter and raises interrupt if
        requested.

        Args:
            message (T): The message from the server to be handled.

        Returns:
            AsyncIterable[Optional[T]]: An asynchronous iterable of reply messages.
        """
        reply, interrupt = await self.interpreter.handle(self, message)
        if interrupt:
            await self.raise_interrupt()
        return reply

    async def run_forever(
        self, device: Device, raise_interrupt: RaiseInterrupt
    ) -> None:
        """Runs the server continously."""
        await super().run_forever(device, raise_interrupt)
        await self.server.run_forever(self.on_connect, self.handle_message)



import pytest

from tickit.core.state_interfaces import state_interface
from tickit.core.state_interfaces.state_interface import (
    StateConsumer,
    StateProducer,
    add,
    get_interface,
    interfaces,
)


@pytest.fixture(autouse=True)
def reset_consumers():
    old_consumers = state_interface.consumers
    state_interface.consumers = dict()
    yield
    state_interface.consumers = old_consumers


@pytest.fixture(autouse=True)
def reset_producers():
    old_producers = state_interface.producers
    state_interface.producers = dict()
    yield
    state_interface.producers = old_producers


@pytest.fixture
def TestStateConsumer():
    return type("TestStateConsumer", (StateConsumer,), dict())


@pytest.fixture
def TestStateProducer():
    return type("TestStateProducer", (StateProducer,), dict())


def test_test_interface_not_in_interfaces():
    assert set() == interfaces()


def test_add_only_consumer_not_in_interfaces(TestStateConsumer):
    add("Test", False)(TestStateConsumer)
    assert "Test" not in interfaces()


def test_add_only_producer_not_in_interfaces(TestStateProducer):
    add("Test", False)(TestStateProducer)
    assert "Test" not in interfaces()


def test_add_both_adds_to_interfaces(TestStateConsumer, TestStateProducer):
    add("Test", False)(TestStateConsumer)
    add("Test", False)(TestStateProducer)
    assert "Test" in interfaces()


def test_add_non_interface_warns():
    with pytest.warns(RuntimeWarning):
        add("Test", False)(type("TestClass", tuple(), dict()))


def test_get_returns_consumer(TestStateConsumer, TestStateProducer):
    add("Test", False)(TestStateConsumer)
    add("Test", False)(TestStateProducer)
    consumer, _ = get_interface("Test")
    assert consumer == TestStateConsumer


def test_get_returns_producer(TestStateConsumer, TestStateProducer):
    add("Test", False)(TestStateConsumer)
    add("Test", False)(TestStateProducer)
    _, producer = get_interface("Test")
    assert producer == TestStateProducer



import pytest

from tickit.devices.eiger.stream.eiger_stream import EigerStream


@pytest.fixture
def filewriter() -> EigerStream:
    return EigerStream()


def test_eiger_stream_constructor():
    EigerStream()



import asyncio
from typing import Iterable

import pytest
from immutables import Map
from mock import Mock, create_autospec, patch

from tickit.core.adapter import Adapter
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import Changes, ComponentID, Output, PortID, SimTime
from tickit.devices.source import SourceDevice


@pytest.fixture
def source() -> SourceDevice:
    source = SourceDevice(value=42)
    return source


@pytest.fixture
def mock_adapter() -> Mock:
    return create_autospec(Adapter)


@pytest.fixture
def mock_state_producer_type() -> Mock:
    mock: Mock = create_autospec(StateProducer, instance=False)
    mock.return_value = create_autospec(StateProducer, instance=True)
    return mock


@pytest.fixture
def mock_state_consumer_type() -> Mock:
    return create_autospec(StateConsumer, instance=False)


@pytest.fixture
def patch_asyncio_wait() -> Iterable[Mock]:
    with patch(
        "tickit.core.components.device_simulation.asyncio.wait", autospec=True
    ) as mock:
        yield mock


@pytest.fixture
def patch_run_all() -> Iterable[Mock]:
    with patch(
        "tickit.core.components.device_simulation.run_all", autospec=True
    ) as mock:
        mock.return_value = asyncio.create_task(asyncio.sleep(0))
        yield mock


@pytest.fixture
def device_simulation(
    source: SourceDevice,
    mock_adapter: Adapter,
) -> DeviceSimulation:
    return DeviceSimulation(
        name=ComponentID("test_device_simulation"),
        device=source,
        adapters=[mock_adapter],
    )


def test_device_simulation_constructor(device_simulation: DeviceSimulation):
    pass


@pytest.mark.asyncio
async def test_device_simulation_run_forever_method(
    device_simulation: DeviceSimulation,
    mock_state_producer_type: Mock,
    mock_state_consumer_type: Mock,
    patch_asyncio_wait: Mock,
):
    with patch(
        "tickit.core.components.device_simulation.run_all", autospec=True
    ) as mock_all:
        mock_all.return_value = [asyncio.create_task(asyncio.sleep(0))]
        await device_simulation.run_forever(
            mock_state_consumer_type, mock_state_producer_type
        )
        patch_asyncio_wait.assert_awaited_once_with(
            mock_all.return_value, return_when=asyncio.FIRST_COMPLETED
        )

    changes = Changes(Map({PortID("foo"): 43}))
    await device_simulation.on_tick(SimTime(1), changes)

    device_simulation.adapters[0].after_update.assert_called_once()  # type: ignore
    assert device_simulation.last_outputs == {"value": 42}
    device_simulation.state_producer.produce.assert_awaited_once_with(  # type: ignore
        "tickit-test_device_simulation-out",
        Output(
            source=ComponentID("test_device_simulation"),
            time=SimTime(1),
            changes=Map({"value": 42}),  # type: ignore
            call_at=None,
        ),
    )



from dataclasses import dataclass, field, fields
from typing import Any

from tickit.devices.eiger.eiger_schema import rw_str


@dataclass
class StreamConfig:
    """Eiger stream configuration taken from the API spec."""

    mode: str = field(
        default="enabled", metadata=rw_str(allowed_values=["disabled", "enabled"])
    )
    header_detail: str = field(
        default="basic", metadata=rw_str(allowed_values=["all", "basic", "none"])
    )
    header_appendix: str = field(default="", metadata=rw_str())
    image_appendix: str = field(default="", metadata=rw_str())

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        f = {}
        for field_ in fields(self):
            f[field_.name] = {
                "value": vars(self)[field_.name],
                "metadata": field_.metadata,
            }
        return f[key]



from tickit.adapters.interpreters.command.command_interpreter import CommandInterpreter
from tickit.adapters.interpreters.command.regex_command import RegexCommand

__all__ = ["CommandInterpreter", "RegexCommand"]



import pytest

from tickit.devices.eiger.filewriter.eiger_filewriter import EigerFileWriter


@pytest.fixture
def filewriter() -> EigerFileWriter:
    return EigerFileWriter()


def test_eiger_filewriter_constructor():
    EigerFileWriter()



from typing import Iterable

import pytest
import pytest_asyncio
from mock import AsyncMock, Mock, patch

from tickit.core.state_interfaces.kafka import KafkaStateConsumer, KafkaStateProducer


@pytest.fixture
def mock_callback() -> AsyncMock:
    async def callback_fn(input: str) -> None:
        return None

    return AsyncMock(spec_set=callback_fn)


@pytest.fixture
def patch_AIOKafkaStateProducer() -> Iterable[Mock]:
    with patch(
        "tickit.core.state_interfaces.kafka.AIOKafkaProducer", autospec=True
    ) as mock:
        yield mock


@pytest.fixture
def patch_AIOKafkaStateConsumer() -> Iterable[Mock]:
    with patch(
        "tickit.core.state_interfaces.kafka.AIOKafkaConsumer", autospec=True
    ) as mock:
        yield mock


@pytest.fixture
def patch_kafka_state_consumer_run_forever_method() -> Iterable[Mock]:
    with patch.object(KafkaStateConsumer, "_run_forever") as mock:
        yield mock


@pytest_asyncio.fixture
async def kafka_state_consumer(
    mock_callback: AsyncMock,
    patch_AIOKafkaStateConsumer: Mock,
    patch_kafka_state_consumer_run_forever_method: Mock,
):
    return KafkaStateConsumer(callback=mock_callback)


@pytest.fixture
def kafka_state_producer(
    mock_callback: AsyncMock,
    patch_AIOKafkaStateProducer: Mock,
):
    return KafkaStateProducer()


@pytest.mark.asyncio
async def test_kafka_state_consumer_constructor(
    kafka_state_consumer: KafkaStateConsumer,
):
    kafka_state_consumer._run_forever.assert_called_once()  # type: ignore
    pass


@pytest.mark.asyncio
async def test_kafka_state_consumer_subscribe(
    kafka_state_consumer: KafkaStateConsumer,
):
    await kafka_state_consumer.subscribe(["foo", "bar"])
    kafka_state_consumer.consumer.subscribe.assert_called_once_with(["foo", "bar"])


@pytest.mark.asyncio
async def test_kafka_state_producer_constructor(
    patch_AIOKafkaStateProducer: Mock,
):
    _: KafkaStateProducer = KafkaStateProducer()


@pytest.mark.asyncio
async def test_kafka_state_producer_produce(
    patch_AIOKafkaStateProducer: Mock,
):
    producer: KafkaStateProducer = KafkaStateProducer()
    await producer.produce("foo", 42)
    producer.producer.send.assert_awaited_once_with("foo", 42)



from typing import AnyStr, AsyncIterable, Iterable


async def wrap_as_async_iterable(message: AnyStr) -> AsyncIterable[AnyStr]:
    """Wraps a message in an asynchronous iterable.

    Args:
        message (AnyStr): A singular message.

    Returns:
        AsyncIterable[AnyStr]: An asynchronous iterable containing the message.
    """
    yield message


async def wrap_messages_as_async_iterable(
    messages: Iterable[AnyStr],
) -> AsyncIterable[AnyStr]:
    """Wraps a message in an asynchronous iterable.

    Args:
        message (AnyStr): An iterable containing a number of messages.

    Returns:
        AsyncIterable[AnyStr]: An asynchronous iterable containing the messages.
    """
    for message in messages:
        yield message



import logging

from aiohttp import web
from apischema import serialize
from typing_extensions import TypedDict

from tickit.adapters.interpreters.endpoints.http_endpoint import HTTPEndpoint
from tickit.core.typedefs import SimTime
from tickit.devices.eiger.eiger_schema import Value
from tickit.devices.eiger.monitor.monitor_config import MonitorConfig
from tickit.devices.eiger.monitor.monitor_status import MonitorStatus

LOGGER = logging.getLogger(__name__)

MONITOR_API = "monitor/api/1.8.0"


class EigerMonitor:
    """Simulation of an Eiger Monitor."""

    #: An empty typed mapping of input values
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: A typed mapping containing the 'value' output value
    Outputs: TypedDict = TypedDict("Outputs", {})

    def __init__(self) -> None:
        """An Eiger Monitor constructor."""
        self.monitor_status: MonitorStatus = MonitorStatus()
        self.monitor_config: MonitorConfig = MonitorConfig()
        self.monitor_callback_period = SimTime(int(1e9))


class EigerMonitorAdapter:
    """An adapter for the Monitor."""

    device: EigerMonitor

    @HTTPEndpoint.get(f"/{MONITOR_API}" + "/config/{param}")
    async def get_monitor_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting config values from the Monitor.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]
        val = self.device.monitor_config[param]["value"]
        meta = self.device.monitor_config[param]["metadata"]

        data = serialize(
            Value(val, meta["value_type"].value, access_mode=meta["access_mode"])
        )

        return web.json_response(data)

    @HTTPEndpoint.get(f"/{MONITOR_API}" + "/status/{param}")
    async def get_monitor_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting status values from the Monitor.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]
        val = self.device.monitor_status[param]["value"]
        meta = self.device.monitor_status[param]["metadata"]

        data = serialize(
            Value(val, meta["value_type"].value, access_mode=meta["access_mode"])
        )

        return web.json_response(data)



from typing import Awaitable, Iterable, Set

import pytest
import pytest_asyncio
from mock import AsyncMock, Mock, create_autospec, patch

from tickit.core.management.event_router import Wiring
from tickit.core.management.schedulers.master import MasterScheduler
from tickit.core.management.ticker import Ticker
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import ComponentID, SimTime


@pytest.fixture
def mock_wiring() -> Mock:
    return create_autospec(Wiring, instance=True)


@pytest.fixture
def mock_state_consumer() -> Mock:
    return create_autospec(StateConsumer, instance=False)


@pytest.fixture
def mock_state_producer() -> Mock:
    return create_autospec(StateProducer, instance=False)


@pytest.fixture
def patch_ticker() -> Iterable[Mock]:
    with patch("tickit.core.management.schedulers.base.Ticker", autospec=True) as mock:
        mock.return_value = AsyncMock(spec=Ticker)
        mock.return_value.time = SimTime(41)
        yield mock


@pytest_asyncio.fixture
async def master_scheduler(
    mock_wiring, mock_state_consumer, mock_state_producer, patch_ticker
) -> MasterScheduler:
    master_scheduler = MasterScheduler(
        mock_wiring, mock_state_consumer, mock_state_producer
    )
    await master_scheduler.setup()
    return master_scheduler


def test_master_scheduler_constructor_and_setup_methods(
    master_scheduler: MasterScheduler,
):
    pass


@pytest.fixture
def patch_asyncio_wait() -> Iterable[Mock]:
    """A patch over the method asyncio.wait.

    This fixture patches over asyncio.wait to prevent the test from waiting
    indefinitely.The method is replaced with a new callable that simulates the
    outcome of the asyncio.wait method in order to cover most of the tested function.
    """

    class wait:
        def __init__(self):
            self.call_count = 0
            self.threshold = 3

        async def __call__(self, awaitables: Set[Awaitable], return_when):
            """Simulates the call to asyncio.wait(return_when=FIRST_COMPLETED).

            The call to wait is a (deliberate) race condition. To simulate this and
            test all parts of the code, this method switches it's response depending
            on how many times it's been called.
            """
            for aw in awaitables:
                await aw

            if self.call_count < self.threshold:
                self.call_count += 1
                return [list(awaitables)[0]], None
            else:
                self.call_count += 1
                return [list(awaitables)[1]], None

    with patch(
        "tickit.core.management.schedulers.master.asyncio.wait",
        new_callable=wait,
    ) as mock:
        yield mock


@pytest.fixture
def patch_asyncio_Event() -> Iterable[Mock]:
    """A patch over the asyncio.Event class. Prevents warnings from being raised."""
    with patch(
        "tickit.core.management.schedulers.master.asyncio.Event", autospec=True
    ) as mock:
        yield mock


@pytest.fixture
def patch_asyncio_sleep() -> Iterable[Mock]:
    with patch(
        "tickit.core.management.schedulers.master.asyncio.sleep", autospec=True
    ) as mock:
        yield mock


@pytest.fixture
def patch_time_ns() -> Iterable[Mock]:
    class frozen_clock:
        def __init__(self):
            self.time = 1000

        def __call__(self):
            return self.time

    with patch(
        "tickit.core.management.schedulers.master.time_ns", new_callable=frozen_clock
    ) as mock:
        yield mock


@pytest.mark.asyncio
async def test_master_scheduler_do_initial_tick(
    master_scheduler: MasterScheduler, patch_time_ns
):
    await master_scheduler.setup()
    await master_scheduler._do_initial_tick()
    assert master_scheduler.last_time == patch_time_ns()


@pytest.mark.asyncio
async def test_master_scheduler_do_tick_method(
    master_scheduler: MasterScheduler,
    patch_asyncio_wait,
    patch_asyncio_Event,
    patch_asyncio_sleep,
):
    master_scheduler.add_wakeup(ComponentID("foo"), SimTime(42))
    master_scheduler.add_wakeup(ComponentID("bar"), SimTime(52))
    master_scheduler.add_wakeup(ComponentID("baz"), SimTime(62))
    master_scheduler.add_wakeup(ComponentID("boo"), SimTime(72))
    await master_scheduler.setup()
    await master_scheduler._do_initial_tick()
    await master_scheduler._do_tick()
    assert master_scheduler.wakeups == {"bar": 52, "baz": 62, "boo": 72}
    await master_scheduler._do_tick()
    assert master_scheduler.wakeups == {"baz": 62, "boo": 72}
    await master_scheduler._do_tick()
    assert master_scheduler.wakeups == {"boo": 72}
    # After two ticks `patch_asyncio_wait` will switch to returning:
    # `self.new_wakeup.wait()` instead of `self.sleep_time(when)` (See:
    # tickit/core/management/schedulers/master.py:89). After that we expect wakeups
    # to stop being removed.
    await master_scheduler._do_tick()
    assert master_scheduler.wakeups == {"boo": 72}


@pytest.mark.asyncio
async def test_master_scheduler_schedule_interrupt_method(
    master_scheduler: MasterScheduler, patch_time_ns
):
    await master_scheduler.setup()
    await master_scheduler._do_initial_tick()
    assert master_scheduler.wakeups == {}
    await master_scheduler.schedule_interrupt(ComponentID("foo"))
    assert master_scheduler.wakeups == {"foo": master_scheduler.ticker.time}


@pytest.mark.asyncio
async def test_schedule_interrupt_succeeds_with_queued_wakeup(
    master_scheduler: MasterScheduler, patch_time_ns
):
    master_scheduler.add_wakeup(ComponentID("bar"), SimTime(42))
    await master_scheduler.setup()
    await master_scheduler._do_initial_tick()
    assert master_scheduler.wakeups == {"bar": 42}
    await master_scheduler.schedule_interrupt(ComponentID("me first"))
    assert master_scheduler.wakeups == {
        "me first": master_scheduler.ticker.time,
        "bar": 42,
    }



from typing import Iterable

import pytest
from aiohttp import web
from mock import Mock
from mock.mock import create_autospec, patch

from tickit.adapters.httpadapter import HTTPAdapter
from tickit.adapters.interpreters.endpoints.http_endpoint import HTTPEndpoint
from tickit.core.device import Device


@pytest.fixture
def mock_device() -> Device:
    return create_autospec(Device)


@pytest.fixture
def mock_raise_interrupt():
    async def raise_interrupt():
        return False

    return Mock(raise_interrupt)


class MockAdapter(HTTPAdapter):
    device: Device

    @HTTPEndpoint.get("/mock_endpoint")
    async def mock_endpoint(self, request: web.Request) -> web.Response:
        return web.Response(text="test")


@pytest.fixture
def http_adapter() -> HTTPAdapter:
    http_adapter = HTTPAdapter()
    return http_adapter


def test_http_adapter_constructor():
    HTTPAdapter()


@pytest.fixture
def patch_asyncio_event_wait() -> Iterable[Mock]:
    with patch(
        "tickit.core.components.device_simulation.asyncio.Event.wait", autospec=True
    ) as mock:
        yield mock


@pytest.mark.asyncio
async def test_http_adapter_run_forever_method(
    http_adapter,
    mock_device: Device,
    mock_raise_interrupt: Mock,
    patch_asyncio_event_wait: Mock,
):
    await http_adapter.run_forever(mock_device, mock_raise_interrupt)

    patch_asyncio_event_wait.assert_awaited_once()


@pytest.mark.asyncio
async def test_http_adapter_endpoints():
    adapter = MockAdapter()

    resp = await list(adapter.endpoints())[0].handler(None)

    assert resp.text == "test"



import json
import logging

from aiohttp import web
from apischema import serialize

from tickit.adapters.httpadapter import HTTPAdapter
from tickit.adapters.interpreters.endpoints.http_endpoint import HTTPEndpoint
from tickit.adapters.zmqadapter import ZeroMQAdapter
from tickit.devices.eiger.eiger import EigerDevice
from tickit.devices.eiger.eiger_schema import AccessMode, SequenceComplete, Value
from tickit.devices.eiger.eiger_status import State
from tickit.devices.eiger.filewriter.eiger_filewriter import EigerFileWriterAdapter
from tickit.devices.eiger.monitor.eiger_monitor import EigerMonitorAdapter
from tickit.devices.eiger.stream.eiger_stream import EigerStreamAdapter

DETECTOR_API = "detector/api/1.8.0"

LOGGER = logging.getLogger(__name__)


class EigerRESTAdapter(
    HTTPAdapter, EigerStreamAdapter, EigerMonitorAdapter, EigerFileWriterAdapter
):
    """An Eiger adapter which parses the commands sent to the HTTP server."""

    device: EigerDevice  # type: ignore

    @HTTPEndpoint.get(f"/{DETECTOR_API}" + "/config/{parameter_name}")
    async def get_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting configuration variables from the Eiger.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["parameter_name"]

        if hasattr(self.device.settings, param):
            attr = self.device.settings[param]

            data = serialize(
                Value(
                    attr["value"],
                    attr["metadata"]["value_type"].value,
                    access_mode=(
                        attr["metadata"]["access_mode"].value  # type: ignore
                        if hasattr(attr["metadata"], "access_mode")
                        else AccessMode.READ_ONLY
                    ),
                )
            )
        else:
            data = serialize(
                Value("None", "string", access_mode=AccessMode.NONE)  # type: ignore
            )

        return web.json_response(data)

    @HTTPEndpoint.put(
        f"/{DETECTOR_API}" + "/config/{parameter_name}", include_json=True
    )
    async def put_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for setting configuration variables for the Eiger.

        Args:
            request (web.Request): The request object that takes the given parameter
            and value.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["parameter_name"]

        response = json.loads(await request.json())

        if self.device.get_state()["value"] != State.IDLE.value:  # type: ignore
            LOGGER.warning("Eiger not initialized or is currently running.")
            return web.json_response(serialize(SequenceComplete(7)))
        elif (
            hasattr(self.device.settings, param)
            and self.device.get_state()["value"] == State.IDLE.value  # type: ignore
        ):
            attr = response["value"]

            LOGGER.debug(f"Changing to {attr} for {param}")

            self.device.settings[param] = attr

            LOGGER.debug("Set " + str(param) + " to " + str(attr))
            return web.json_response(serialize(SequenceComplete(8)))
        else:
            LOGGER.debug("Eiger has no config variable: " + str(param))
            return web.json_response(serialize(SequenceComplete(9)))

    @HTTPEndpoint.get(f"/{DETECTOR_API}" + "/status/{status_param}")
    async def get_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting the status of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["status_param"]

        if hasattr(self.device.status, param):
            attr = self.device.status[param]
        else:
            attr = "None"

        data = serialize({"value": attr})

        return web.json_response(data)

    @HTTPEndpoint.get(f"/{DETECTOR_API}" + "/status/board_000/{status_param}")
    async def get_board_000_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting the status of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["status_param"]

        if hasattr(self.device.status, param):
            attr = self.device.status[param]
        else:
            attr = "None"

        data = serialize({"value": attr})

        return web.json_response(data)

    @HTTPEndpoint.get(f"/{DETECTOR_API}" + "/status/builder/{status_param}")
    async def get_builder_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting the status of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["status_param"]

        if hasattr(self.device.status, param):
            attr = self.device.status[param]
        else:
            attr = "None"

        data = serialize({"value": attr})

        return web.json_response(data)

    @HTTPEndpoint.put(f"/{DETECTOR_API}" + "/command/initialize")
    async def initialize_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'initialize' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        await self.device.initialize()

        LOGGER.debug("Initializing Eiger...")
        return web.json_response(serialize(SequenceComplete(1)))

    @HTTPEndpoint.put(f"/{DETECTOR_API}" + "/command/arm")
    async def arm_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'arm' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        await self.device.arm()

        LOGGER.debug("Arming Eiger...")
        return web.json_response(serialize(SequenceComplete(2)))

    @HTTPEndpoint.put(f"/{DETECTOR_API}" + "/command/disarm")
    async def disarm_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'disarm' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        await self.device.disarm()

        LOGGER.debug("Disarming Eiger...")
        return web.json_response(serialize(SequenceComplete(3)))

    @HTTPEndpoint.put(f"/{DETECTOR_API}" + "/command/trigger")
    async def trigger_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'trigger' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        trigger_message = await self.device.trigger()
        self.device._set_state(State.IDLE)

        LOGGER.debug(trigger_message)
        return web.json_response(serialize(SequenceComplete(4)))

    @HTTPEndpoint.put(f"/{DETECTOR_API}" + "/command/cancel")
    async def cancel_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'cancel' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        await self.device.cancel()

        LOGGER.debug("Cancelling Eiger...")
        return web.json_response(serialize(SequenceComplete(5)))

    @HTTPEndpoint.put(f"/{DETECTOR_API}" + "/command/abort")
    async def abort_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'abort' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        await self.device.abort()

        LOGGER.debug("Aborting Eiger...")
        return web.json_response(serialize(SequenceComplete(6)))


class EigerZMQAdapter(ZeroMQAdapter):
    """An Eiger adapter which parses the data to send along a ZeroMQStream."""

    device: EigerDevice



from collections import defaultdict

import pytest
from mock import MagicMock

from tickit.core.state_interfaces.internal import (
    InternalStateConsumer,
    InternalStateProducer,
    InternalStateServer,
    Message,
    Messages,
)
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.utils.singleton import Singleton


@pytest.fixture(autouse=True)
def reset_topics():
    InternalStateServer()._topics = defaultdict(lambda: Messages(list()))


@pytest.fixture
def internal_state_server():
    return InternalStateServer()


@pytest.fixture
def internal_state_consumer() -> InternalStateConsumer:
    return InternalStateConsumer(MagicMock())


@pytest.fixture
def internal_state_producer():
    return InternalStateProducer()


@pytest.fixture
def mock_async_callback():
    async def async_callback(value: object) -> None:
        pass

    return MagicMock(async_callback)


def test_internal_state_server_is_singleton():
    assert isinstance(InternalStateServer, Singleton)


def test_internal_state_server_topics_empty(internal_state_server: InternalStateServer):
    assert not internal_state_server._topics


def test_internal_state_server_create_topic_creates_topic(
    internal_state_server: InternalStateServer,
):
    internal_state_server.create_topic("Test")
    assert "Test" in internal_state_server.topics


def test_internal_state_server_remove_topic_removes_topic(
    internal_state_server: InternalStateServer,
):
    internal_state_server.create_topic("Test")
    internal_state_server.remove_topic("Test")
    assert "Test" not in internal_state_server.topics


def test_internal_state_server_create_topic_creates_empty(
    internal_state_server: InternalStateServer,
):
    internal_state_server.create_topic("Test")
    assert list() == internal_state_server._topics["Test"]


@pytest.mark.asyncio
async def test_internal_state_server_push_adds_message(
    internal_state_server: InternalStateServer,
):
    internal_state_server.create_topic("Test")
    await internal_state_server.push("Test", Message("TestMessage"))
    assert Message("TestMessage") in internal_state_server._topics["Test"]


def test_internal_state_consumer_is_state_consumer():
    assert isinstance(InternalStateConsumer, StateConsumer)


@pytest.mark.asyncio
async def test_internal_state_consumer_consumes(
    internal_state_server: InternalStateServer, mock_async_callback: MagicMock
):
    internal_state_consumer = InternalStateConsumer(mock_async_callback)
    await internal_state_consumer.subscribe(["Test"])
    await internal_state_server.push("Test", Message("TestMessage"))
    mock_async_callback.assert_called_once_with("TestMessage")


@pytest.mark.asyncio
async def test_internal_state_consumer_consumes_prior(
    internal_state_server: InternalStateServer, mock_async_callback: MagicMock
):
    internal_state_consumer = InternalStateConsumer(mock_async_callback)
    await internal_state_server.push("Test", Message("TestMessage"))
    await internal_state_consumer.subscribe(["Test"])
    mock_async_callback.assert_called_once_with("TestMessage")


@pytest.mark.asyncio
async def test_internal_state_consumer_consumes_multiple(
    internal_state_server: InternalStateServer, mock_async_callback: MagicMock
):
    internal_state_consumer = InternalStateConsumer(mock_async_callback)
    await internal_state_consumer.subscribe(["Test"])
    await internal_state_server.push("Test", Message("TestMessage"))
    mock_async_callback.assert_called_with("TestMessage")
    await internal_state_server.push("Test", Message("OtherTestMessage"))
    mock_async_callback.assert_called_with("OtherTestMessage")


def test_internal_state_producer_is_state_producer():
    assert isinstance(InternalStateProducer, StateProducer)


@pytest.mark.asyncio
async def test_internal_state_producer_produces(
    internal_state_server: InternalStateServer,
    internal_state_producer: InternalStateProducer,
):
    await internal_state_producer.produce("Test", "TestMessage")
    assert Message("TestMessage") in internal_state_server._topics["Test"]



import pytest
from aiohttp import web

from tickit.adapters.interpreters.endpoints.http_endpoint import HTTPEndpoint


@pytest.fixture
def http_endpoint(url: str, method: str, include_json: bool, interrupt: bool):
    return HTTPEndpoint(url, method, include_json, interrupt)


def test_http_endpoint_registers_get_endpoint():
    class TestAdapter:
        @HTTPEndpoint.get("test", False, False)
        def test_endpoint():
            pass

    assert isinstance(TestAdapter.test_endpoint.__endpoint__, HTTPEndpoint)


def test_http_endpoint_registers_put_endpoint():
    class TestAdapter:
        @HTTPEndpoint.put("test", False, False)
        def test_endpoint():
            pass

    assert isinstance(TestAdapter.test_endpoint.__endpoint__, HTTPEndpoint)


def fake_endpoint(request: web.Request):
    pass


@pytest.mark.parametrize(
    ["url", "method", "include_json", "interrupt", "expected"],
    [
        (
            r"TestUrl",
            r"GET",
            False,
            False,
            web.RouteDef("GET", "TestUrl", fake_endpoint, {}),
        )
    ],
)
def test_http_get_endpoint_define_returns_get_routedef(
    http_endpoint: HTTPEndpoint, expected: web.RouteDef
):
    assert expected == http_endpoint.define(fake_endpoint)


@pytest.mark.parametrize(
    ["url", "method", "include_json", "interrupt", "expected"],
    [
        (
            r"TestUrl",
            r"PUT",
            False,
            False,
            web.RouteDef("PUT", "TestUrl", fake_endpoint, {}),
        )
    ],
)
def test_http_put_endpoint_define_returns_put_routedef(
    http_endpoint: HTTPEndpoint, expected: web.RouteDef
):
    assert expected == http_endpoint.define(fake_endpoint)



from dataclasses import dataclass
from random import random
from typing import Optional

from tickit.adapters.composed import ComposedAdapter
from tickit.adapters.interpreters.command.command_interpreter import CommandInterpreter
from tickit.adapters.interpreters.command.regex_command import RegexCommand
from tickit.adapters.servers.tcp import TcpServer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.byte_format import ByteFormat
from tickit.utils.compat.typing_compat import TypedDict


class ShutterDevice(Device):
    """A toy device which downscales flux according to a set position.

    A toy device which produces an output flux which is downscaled from the input flux
    according to an internal state position. The position may be altered by setting new
    a target which will be matched by position over a period of time determined by the
    rate.
    """

    #: A typed mapping containing the 'flux' input value
    Inputs: TypedDict = TypedDict("Inputs", {"flux": float})
    #: A typed mapping containing the 'flux' output value
    Outputs: TypedDict = TypedDict("Outputs", {"flux": float})

    def __init__(
        self, default_position: float, initial_position: Optional[float] = None
    ) -> None:
        """A Shutter constructor which configures the initial and default position.

        Args:
            default_position (float): The initial target position of the shutter
            initial_position (Optional[float]): The initial position of the shutter. If
                None, a random value in the range [0.0,1.0) will be used. Defaults to
                None.
        """
        self.target_position = default_position
        self.position = initial_position if initial_position else random()
        self.rate = 2e-10
        self.last_time: Optional[SimTime] = None

    @staticmethod
    def move(position: float, target: float, rate: float, period: SimTime) -> float:
        """A helper method used to compute the new position of a shutter.

        A helper method used to compute the new position of a shutter given a target
        position, a rate of change and a period over which the change occurs. Movement
        is performed at the defined rate and comes to a "hard" stop when the desired
        position is reached.

        Args:
            position (float): The prior position of the shutter.
            target (float): The target position of the shutter.
            rate (float): The rate of change of shutter position.
            period (SimTime): The period over which the change occurs.

        Returns:
            float: The posterior position of the shutter.
        """
        if position < target:
            position = min(position + rate * period, target)
        elif position > target:
            position = max(position - rate * period, target)
        return position

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which moves the shutter and produces a downscaled flux.

        The update method which adjusts the position according to the target position,
        computes the transmitted flux and produces the output flux with a request to be
        called back in 100ms if the if the shutter continues to move.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the transmitted
                flux, and requests a callback after 100ms if the shutter continues to
                move.
        """
        if self.last_time:
            self.position = self.move(
                self.position,
                self.target_position,
                self.rate,
                SimTime(time - self.last_time),
            )
        self.last_time = time
        call_at = (
            None if self.position == self.target_position else SimTime(time + int(1e8))
        )
        output_flux = inputs["flux"] * self.position
        return DeviceUpdate(self.Outputs(flux=output_flux), call_at)


class ShutterAdapter(ComposedAdapter):
    """A toy composed adapter which gets shutter position and target and sets target."""

    device: ShutterDevice

    def __init__(
        self,
        host: str = "localhost",
        port: int = 25565,
    ) -> None:
        """A Shutter which instantiates a TcpServer with configured host and port.

        Args:
            device (Device): The device which this adapter is attached to
            raise_interrupt (Callable): A callback to request that the device is
                updated immediately.
            host (Optional[str]): The host address of the TcpServer. Defaults to
                "localhost".
            port (Optional[int]): The bound port of the TcpServer. Defaults to 25565.
        """
        super().__init__(
            TcpServer(host, port, ByteFormat(b"%b\r\n")),
            CommandInterpreter(),
        )

    @RegexCommand(r"P\?", False, "utf-8")
    async def get_position(self) -> bytes:
        """A regex string command which returns the utf-8 encoded value of position.

        Returns:
            bytes: The utf-8 encoded value of position.
        """
        return str(self.device.position).encode("utf-8")

    @RegexCommand(r"T\?", False, "utf-8")
    async def get_target(self) -> bytes:
        """A regex string command which returns the utf-8 encoded value of target.

        Returns:
            bytes: The utf-8 encoded value of target.
        """
        return str(self.device.target_position).encode("utf-8")

    @RegexCommand(r"T=(\d+\.?\d*)", True, "utf-8")
    async def set_target(self, target: str) -> None:
        """A regex string command which sets the target position of the shutter.

        Args:
            target (str): The target position of the shutter.
        """
        self.device.target_position = float(target)
        self.device.last_time = None


@dataclass
class Shutter(ComponentConfig):
    """Shutter you can open or close over TCP."""

    default_position: float
    initial_position: Optional[float] = None
    host: str = "localhost"
    port: int = 25565

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=ShutterDevice(
                default_position=self.default_position,
                initial_position=self.initial_position,
            ),
            adapters=[ShutterAdapter(host=self.host, port=self.port)],
        )



import pytest

from tickit.devices.eiger.monitor.monitor_status import MonitorStatus

# # # # # Eiger MonitorStatus Tests # # # # #


@pytest.fixture
def monitor_status() -> MonitorStatus:
    return MonitorStatus()


def test_eiger_monitor_status_constructor():
    MonitorStatus()


def test_eiger_monitor_status_getitem(monitor_status):
    assert [] == monitor_status["error"]["value"]



import asyncio
from typing import Awaitable, Callable, Generic, Iterable, TypeVar

import yaml
from aiokafka import AIOKafkaConsumer, AIOKafkaProducer
from yaml.loader import Loader

from tickit.core.state_interfaces import state_interface

#: A consumable value
C = TypeVar("C")
#: A producable value
P = TypeVar("P")


@state_interface.add("kafka", True)
class KafkaStateConsumer(Generic[C]):
    """A kafka implementation of the StateConsumer protocol.

    A kafka implementation of the StateConsumer protocol, this consumer can subscribe
    to kafka topics, upon recieving a message the consumer passes the value to the
    callback function passed during initialization, if a topic is subscribed to which
    does not yet exist it is created.
    """

    def __init__(self, callback: Callable[[C], Awaitable[None]]) -> None:
        """Creates an AIOKafka Consumer and begins consuming subscribed topics.

        Args:
            callback (Callable[[C], Awaitable[None]]): An asynchronous handler function
                for consumed values.
        """
        self.consumer = AIOKafkaConsumer(
            None,
            auto_offset_reset="earliest",
            value_deserializer=lambda m: yaml.load(m.decode("utf-8"), Loader=Loader),
        )
        self.callback = callback
        asyncio.create_task(self._run_forever())

    async def _run_forever(self) -> None:
        """Starts the consumer and waits for messages to arrive."""
        await self.consumer.start()
        while True:
            async for message in self.consumer:
                await self.callback(message.value)

    async def subscribe(self, topics: Iterable[str]):
        """Subscribes the consumer to the given topics, new messages are passed to the
        callback.

        Args:
            topics (Iterable[str]): An iterable of topics to subscribe to.
        """
        self.consumer.subscribe(topics)


@state_interface.add("kafka", True)
class KafkaStateProducer(Generic[P]):
    """A kafka implementation of the StateProducer protocol.

    A kafka implementation of the StateProducer protocol, this producer can produce a
    value to a kafka topic, if the topic does not yet exist it is created.
    """

    def __init__(self) -> None:
        """Creates and starts and AIOKafka Producer."""
        self.producer = AIOKafkaProducer(
            value_serializer=lambda m: yaml.dump(m).encode("utf-8")
        )
        self._start = asyncio.create_task(self.producer.start())

    async def produce(self, topic: str, value: P) -> None:
        """Produces a value to the provided topic.

        Args:
            topic (str): The topic to which the value should be sent.
            value (P): The value to send to the provided topic.
        """
        await self._start
        await self.producer.send(topic, value)



from asyncio import StreamReader, StreamWriter
from typing import AsyncIterable, Awaitable, Callable, Optional

import pytest
from mock import ANY, MagicMock, Mock, patch
from mock.mock import AsyncMock, create_autospec

from tickit.adapters.servers.tcp import TcpServer


@pytest.fixture
def tcp_server() -> TcpServer:
    return TcpServer()


@pytest.fixture
def on_connect() -> Callable[[], AsyncIterable[bytes]]:
    async def on_connect():
        yield b"hello"

    return on_connect


@pytest.fixture
def handler() -> Callable[[bytes], Awaitable[AsyncIterable[bytes]]]:
    async def _handler(message: bytes) -> AsyncIterable[bytes]:
        async def async_iterable() -> AsyncIterable[bytes]:
            yield b"hello"
            yield b""
            yield b"blue"
            yield b"world"
            yield b""

        return async_iterable()

    return _handler


@pytest.fixture
def patch_start_server():
    with patch("asyncio.start_server") as mock:
        yield mock


@pytest.fixture
def mock_stream_reader() -> Mock:
    mock: MagicMock = create_autospec(StreamReader, instance=True)
    mock.read = AsyncMock(side_effect=(b"hello", b"world", b""))
    return mock


@pytest.fixture
def mock_stream_writer() -> Mock:
    mock: MagicMock = create_autospec(StreamWriter, instance=True)
    mock.is_closing = Mock(side_effect=[False, False, True])
    mock.write = Mock()
    mock.drain = AsyncMock()
    return mock


@pytest.mark.asyncio
async def test_TcpServer_run_forever_method(
    tcp_server: TcpServer,
    on_connect: Callable[[], AsyncIterable[bytes]],
    handler: Callable[[bytes], Awaitable[AsyncIterable[Optional[bytes]]]],
    patch_start_server: Mock,
):
    mock_start_server = patch_start_server
    mock_start_server.return_value = AsyncMock()

    await tcp_server.run_forever(on_connect, handler)

    mock_start_server.assert_awaited_once_with(ANY, tcp_server.host, tcp_server.port)
    mock_start_server.return_value.serve_forever.assert_awaited_once()


@pytest.mark.asyncio
async def test_TcpServer_generate_handle_function_method(
    tcp_server: TcpServer,
    on_connect: Callable[[], AsyncIterable[bytes]],
    handler: Callable[[bytes], Awaitable[AsyncIterable[Optional[bytes]]]],
    mock_stream_reader: Mock,
    mock_stream_writer: Mock,
):
    handle_function = tcp_server._generate_handle_function(on_connect, handler)

    assert handle_function is not None

    await handle_function(mock_stream_reader, mock_stream_writer)



from dataclasses import is_dataclass
from typing import Type

import pytest
from immutables import Map
from mock import AsyncMock, create_autospec

from tickit.core.components.component import BaseComponent, ComponentConfig
from tickit.core.state_interfaces.internal import (
    InternalStateConsumer,
    InternalStateProducer,
)
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import Changes, ComponentID, Input, Interrupt, Output, SimTime
from tickit.utils.topic_naming import input_topic, output_topic


def test_component_config_is_dataclass():
    assert is_dataclass(ComponentConfig)


def test_base_component_initialises():
    assert BaseComponent(ComponentID("TestBase"))


@pytest.fixture
def MockConsumer():
    return create_autospec(InternalStateConsumer, instance=False)


@pytest.fixture
def MockProducer():
    return create_autospec(InternalStateProducer, instance=False)


@pytest.fixture
def TestComponent():
    return type("TestComponent", (BaseComponent,), dict())


@pytest.fixture
def test_component(TestComponent: Type[BaseComponent]):
    return TestComponent(ComponentID("TestBase"))  # type: ignore


@pytest.mark.asyncio
async def test_base_component_handle_input_awaits_on_tick(
    test_component: BaseComponent,
):
    test_component.on_tick = AsyncMock()  # type: ignore
    await test_component.handle_input(
        Input(ComponentID("Test"), SimTime(42), Changes(Map()))
    )
    test_component.on_tick.assert_awaited_once_with(SimTime(42), Changes(Map()))


@pytest.mark.asyncio
async def test_base_component_output_sends_output(
    test_component: BaseComponent,
    MockConsumer: Type[StateConsumer],
    MockProducer: Type[StateProducer],
):
    await test_component.run_forever(MockConsumer, MockProducer)
    test_component.state_producer.produce = AsyncMock()  # type: ignore
    await test_component.output(SimTime(42), Changes(Map()), None)
    test_component.state_producer.produce.assert_awaited_once_with(
        output_topic(ComponentID("TestBase")),
        Output(ComponentID("TestBase"), SimTime(42), Changes(Map()), None),
    )


@pytest.mark.asyncio
async def test_base_component_raise_interrupt_sends_output(
    test_component: BaseComponent,
    MockConsumer: Type[StateConsumer],
    MockProducer: Type[StateProducer],
):
    await test_component.run_forever(MockConsumer, MockProducer)
    test_component.state_producer.produce = AsyncMock()  # type: ignore
    await test_component.raise_interrupt()
    test_component.state_producer.produce.assert_awaited_once_with(
        output_topic(ComponentID("TestBase")),
        Interrupt(ComponentID("TestBase")),
    )


@pytest.mark.asyncio
async def test_base_component_set_up_state_interfaces_creates_consumer(
    TestComponent: Type[BaseComponent],
    MockConsumer: Type[StateConsumer],
    MockProducer: Type[StateProducer],
):
    test_component = TestComponent(ComponentID("TestBase"))  # type: ignore
    await test_component.run_forever(MockConsumer, MockProducer)
    assert test_component.state_consumer == MockConsumer(AsyncMock())


@pytest.mark.asyncio
async def test_base_component_set_up_state_interfaces_subscribes_consumer(
    test_component: BaseComponent,
    MockConsumer: Type[StateConsumer],
    MockProducer: Type[StateProducer],
):
    await test_component.run_forever(MockConsumer, MockProducer)
    test_component.state_consumer.subscribe.assert_called_once_with(  # type: ignore
        [input_topic(ComponentID("TestBase"))]
    )


@pytest.mark.asyncio
async def test_base_component_set_up_state_interfaces_creates_producer(
    TestComponent: Type[BaseComponent],
    MockConsumer: Type[StateConsumer],
    MockProducer: Type[StateProducer],
):
    test_component = TestComponent(
        ComponentID("TestBase"),
    )  # type: ignore
    await test_component.run_forever(MockConsumer, MockProducer)
    assert test_component.state_producer == MockProducer()


@pytest.mark.asyncio
async def test_base_component_on_tick_raises_not_implemented(
    test_component: BaseComponent,
):
    with pytest.raises(NotImplementedError):
        await test_component.on_tick(SimTime(42), Changes(Map()))



from dataclasses import dataclass, field, fields
from datetime import datetime
from enum import Enum
from typing import Any, List


class State(Enum):
    """Possible states of the Eiger detector."""

    NA = "na"
    READY = "ready"
    INITIALIZE = "initialize"
    CONFIGURE = "configure"
    ACQUIRE = "acquire"
    IDLE = "idle"
    TEST = "test"
    ERROR = "error"


@dataclass
class EigerStatus:
    """Stores the status parameters of the Eiger detector."""

    state: State = field(default=State.NA)
    errors: List[str] = field(default_factory=list)
    th0_temp: float = field(default=24.5)
    th0_humidity: float = field(default=0.2)
    time: datetime = field(default=datetime.now())
    dcu_buffer_free: float = field(default=0.5)

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        f = {}
        for field_ in fields(self):
            f[field_.name] = vars(self)[field_.name]
        return f[key]



from contextlib import nullcontext as does_not_raise
from random import getrandbits
from typing import Any, Dict

import pytest
from mock import Mock

from tickit.devices.cryostream.base import CryostreamBase
from tickit.devices.cryostream.states import AlarmCodes, PhaseIds, RunModes
from tickit.devices.cryostream.status import ExtendedStatus, Status


def rand_bool() -> bool:
    return bool(getrandbits(1))


@pytest.mark.asyncio
async def test_restart():
    cryostream_base = CryostreamBase()
    await cryostream_base.restart()
    assert cryostream_base.run_mode in (
        RunModes.STARTUPOK.value,
        RunModes.STARTUPFAIL.value,
    )


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "starting_temperature": 10000,
                "target_temperature": 9000,
                "ramp_rate": 360,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.NO_ERRORS.value,
                "expected_gas_flow": 5,
            },
            id="normal params",
        ),
        pytest.param(
            {
                "starting_temperature": 8999,
                "target_temperature": 8000,
                "ramp_rate": 360,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.NO_ERRORS.value,
                "expected_gas_flow": 10,
            },
            id="gas gas gas",
        ),
        pytest.param(
            {
                "starting_temperature": 10000,
                "target_temperature": 90001,
                "ramp_rate": 360,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.TEMP_CONTROL_ERROR.value,
                "expected_gas_flow": 10,
            },
            id="target temperature too high",
        ),
        pytest.param(
            {
                "starting_temperature": 10000,
                "target_temperature": 9000,
                "ramp_rate": 361,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.TEMP_CONTROL_ERROR.value,
                "expected_gas_flow": 5,
            },
            id="ramp rate too high",
        ),
    ],
)
@pytest.mark.asyncio
async def test_ramp(test_params):
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["starting_temperature"]

    await cryostream_base.ramp(
        ramp_rate=test_params["ramp_rate"],
        target_temp=test_params["target_temperature"],
    )

    assert cryostream_base.run_mode == RunModes.RUN.value
    assert cryostream_base.alarm_code == test_params["expected_alarm_code"]
    assert cryostream_base.phase_id == test_params["expected_phase_id"]
    assert cryostream_base.gas_flow == test_params["expected_gas_flow"]


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "duration": 1440,
                "expected_phase_id": PhaseIds.PLAT.value,
                "expected_run_mode": RunModes.RUN.value,
                "raises": does_not_raise(),
            },
            id="normal params",
        ),
        pytest.param(
            {
                "duration": 0,
                "expected_phase_id": None,
                "expected_run_mode": None,
                "raises": pytest.raises(ValueError),
            },
            id="zero duration",
        ),
        pytest.param(
            {
                "duration": 1441,
                "expected_phase_id": None,
                "expected_run_mode": None,
                "raises": pytest.raises(ValueError),
            },
            id="too long",
        ),
    ],
)
@pytest.mark.asyncio
async def test_plat(test_params):
    cryostream_base = CryostreamBase()
    initial_run_mode = cryostream_base.run_mode
    initial_phase_id = cryostream_base.phase_id
    with test_params["raises"]:
        await cryostream_base.plat(test_params["duration"])

    if test_params["expected_phase_id"] is None:
        assert cryostream_base.phase_id == initial_phase_id
    else:
        assert cryostream_base.phase_id == test_params["expected_phase_id"]

    if test_params["expected_run_mode"] is None:
        assert cryostream_base.run_mode == initial_run_mode
    else:
        assert cryostream_base.run_mode == test_params["expected_run_mode"]


@pytest.mark.asyncio
async def test_hold():
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = 28000
    await cryostream_base.hold()
    final_temperature = cryostream_base.update_temperature(100e9)
    assert final_temperature == 28000
    assert cryostream_base.run_mode == RunModes.RUN.value
    assert cryostream_base.phase_id == PhaseIds.HOLD.value


@pytest.mark.asyncio
async def test_cool():
    cryostream_base = CryostreamBase()
    cryostream_base.ramp = Mock(cryostream_base.ramp)
    starting_temperature = cryostream_base.gas_temp
    target_temperature = starting_temperature - 10 * 5
    await cryostream_base.cool(target_temperature)
    assert cryostream_base.run_mode == RunModes.RUN.value
    assert cryostream_base.phase_id == PhaseIds.COOL.value
    cryostream_base.ramp.assert_awaited_once_with(
        cryostream_base.default_ramp_rate, target_temperature
    )


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown,
                "expected_run_mode": RunModes.SHUTDOWNOK.value,
                "expected_gas_flow": {0},
            },
            id="shutdown okay",
        ),
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown - 100,
                "expected_run_mode": RunModes.SHUTDOWNFAIL.value,
                "expected_gas_flow": {5, 10},
            },
            id="shutdown fails",
        ),
    ],
)
@pytest.mark.asyncio
async def test_end(test_params: Dict[str, Any]):
    def set_gas_flow(value: int, target_temp: int) -> None:
        if rand_bool():
            cryostream_base.gas_flow = 10
        else:
            cryostream_base.gas_flow = 5

    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["initial_gas_temp"]
    cryostream_base.ramp = Mock(  # type: ignore
        cryostream_base.ramp, side_effect=set_gas_flow
    )
    await cryostream_base.end(ramp_rate=cryostream_base.max_rate)
    assert cryostream_base.phase_id == PhaseIds.END.value
    assert cryostream_base.run_mode == test_params["expected_run_mode"]
    assert cryostream_base.gas_flow in test_params["expected_gas_flow"]


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown,
                "expected_run_mode": RunModes.SHUTDOWNOK.value,
            },
            id="shutdown okay",
        ),
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown - 100,
                "expected_run_mode": RunModes.SHUTDOWNFAIL.value,
            },
            id="shutdown fails",
        ),
    ],
)
@pytest.mark.asyncio
async def test_purge(test_params: Dict[str, Any]):
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["initial_gas_temp"]
    cryostream_base.ramp = Mock(cryostream_base.ramp)  # type: ignore
    await cryostream_base.purge()
    assert cryostream_base.phase_id == PhaseIds.PURGE.value
    assert cryostream_base.run_mode == test_params["expected_run_mode"]


@pytest.mark.asyncio
async def test_stop():
    cryostream_base = CryostreamBase()
    await cryostream_base.stop()
    assert cryostream_base.gas_flow == 0
    assert cryostream_base._target_temp == cryostream_base.gas_temp
    assert cryostream_base.run_mode == RunModes.SHUTDOWNOK.value


@pytest.mark.parametrize(
    "test_params",
    argvalues=[
        {"temp": 350, "expected_gas_flow": 5},
        {"temp": 300, "expected_gas_flow": 10},
    ],
)
@pytest.mark.asyncio
async def test_turbo(test_params):
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["temp"]
    await cryostream_base.turbo(1)
    assert cryostream_base.turbo_mode == 1
    assert cryostream_base.gas_flow == test_params["expected_gas_flow"]


def test_update_temperature():
    cryostream_base = CryostreamBase()
    cryostream_base._target_temp = cryostream_base.gas_temp - cryostream_base.gas_flow

    cryostream_base.update_temperature(1e9)
    assert cryostream_base.gas_temp == cryostream_base._target_temp


@pytest.mark.asyncio
async def test_set_status_format():
    cryostream_base = CryostreamBase()
    await cryostream_base.set_status_format(0)
    assert isinstance(cryostream_base.status, Status)

    await cryostream_base.set_status_format(1)
    assert isinstance(cryostream_base.extended_status, ExtendedStatus)


@pytest.mark.asyncio
async def test_get_status():
    cryostream_base = CryostreamBase()
    status = await cryostream_base.get_status(0)
    assert isinstance(status, Status)

    extended_status = await cryostream_base.get_status(1)
    assert isinstance(extended_status, ExtendedStatus)



import asyncio
import logging
from asyncio.streams import StreamReader, StreamWriter
from typing import AsyncIterable, Awaitable, Callable, List, Optional

from tickit.core.adapter import Server
from tickit.utils.byte_format import ByteFormat

LOGGER = logging.getLogger(__name__)


class TcpServer(Server[bytes]):
    """A configurable tcp server with delegated message handling for use in adapters."""

    def __init__(
        self,
        host: str = "localhost",
        port: int = 25565,
        format: ByteFormat = ByteFormat(b"%b"),
    ) -> None:
        """The TcpServer constructor which takes a host, port and format byte string.

        Args:
            host (str): The host name which the server should be run under.
            port (int): The port number which the server should listen to.
            format (ByteFormat): A formatting string for messages sent by the server,
                allowing for the prepending and appending of data. Defaults to b"%b".
        """
        self.host = host
        self.port = port
        self.format = format.format

    async def run_forever(
        self,
        on_connect: Callable[[], AsyncIterable[Optional[bytes]]],
        handler: Callable[[bytes], Awaitable[AsyncIterable[Optional[bytes]]]],
    ) -> None:
        """Runs the TCP server indefinitely on the configured host and port.

        An asynchronous method used to run the server indefinitely on the configured
        host and port. Upon client connection, messages from the on_connect iterable
        will be sent. Upon recieving a message the server will delegate handling of it
        to the handler. Replies will be formatted according to the configured format
        string.

        Args:
            on_connect (Callable[[], AsyncIterable[bytes]]): An asynchronous iterable
                of messages to be sent upon client connection.
            handler (Callable[[bytes], Awaitable[AsyncIterable[bytes]]]): An
                asynchronous message handler which returns an asynchronous iterable of
                replies.
        """
        handle = self._generate_handle_function(on_connect, handler)

        server = await asyncio.start_server(handle, self.host, self.port)

        async with server:
            await server.serve_forever()

    def _generate_handle_function(
        self,
        on_connect: Callable[[], AsyncIterable[Optional[bytes]]],
        handler: Callable[[bytes], Awaitable[AsyncIterable[Optional[bytes]]]],
    ) -> Callable[[StreamReader, StreamWriter], Awaitable[None]]:
        """Generates the handle function to be passed to the server.

        The handle function is generated from the specified functions. It's purpose is
        to define how the server will respond to incoming messages.

        Args:
            on_connect (Callable[[], AsyncIterable[bytes]]): An asynchronous iterable
                of messages to be sent upon client connection.
            handler (Callable[[bytes], Awaitable[AsyncIterable[bytes]]]): An
                asynchronous message handler which returns an asynchronous iterable of
                replies.
        """
        tasks: List[asyncio.Task] = list()

        async def handle(reader: StreamReader, writer: StreamWriter) -> None:
            async def reply(replies: AsyncIterable[Optional[bytes]]) -> None:
                async for reply in replies:
                    if reply is None:
                        continue
                    LOGGER.debug("Replying with {!r}".format(reply))
                    writer.write(self.format % reply)
                    if writer.is_closing():
                        break
                    await writer.drain()

            tasks.append(asyncio.create_task(reply(on_connect())))

            while True:
                data: bytes = await reader.read(1024)
                if data == b"":
                    break
                addr = writer.get_extra_info("peername")

                LOGGER.debug("Recieved {!r} from {}".format(data, addr))
                tasks.append(asyncio.create_task(reply(await handler(data))))

            await asyncio.wait(tasks)

        return handle



from dataclasses import dataclass
from typing import Optional

from aiohttp import web

from tickit.adapters.httpadapter import HTTPAdapter
from tickit.adapters.interpreters.endpoints.http_endpoint import HTTPEndpoint
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.compat.typing_compat import TypedDict


class ExampleHTTPDevice(Device):
    """A device class for an example HTTP device.

    ...
    """

    Inputs: TypedDict = TypedDict("Inputs", {"foo": bool})

    Outputs: TypedDict = TypedDict("Outputs", {"bar": float})

    def __init__(
        self,
        foo: bool = False,
        bar: Optional[int] = 10,
    ) -> None:
        """An example HTTP device constructor which configures the ... .

        Args:
            foo (bool): A flag to indicate something. Defauls to False.
            bar (int, optional): A number to represent something. Defaults to 3600.
        """
        self.foo = foo
        self.bar = bar

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """Generic update function to update the values of the ExampleHTTPDevice.

        Args:
            time (SimTime): The simulation time in nanoseconds.
            inputs (Inputs): A TypedDict of the inputs to the ExampleHTTPDevice.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the device
                variables.
        """
        pass


class ExampleHTTPAdapter(HTTPAdapter):
    """An Eiger adapter which parses the commands sent to the HTTP server."""

    device: ExampleHTTPDevice

    @HTTPEndpoint.put("/command/foo/")
    async def foo(self, request: web.Request) -> web.Response:
        """A HTTP endpoint for sending a command to the example HTTP device.

        Args:
            request (web.Request): [description]

        Returns:
            web.Response: [description]
        """
        return web.Response(text=str("put data"))

    @HTTPEndpoint.get("/info/bar/{data}")
    async def bar(self, request: web.Request) -> web.Response:
        """A HTTP endpoint for requesting data from the example HTTP device.

        Args:
            request (web.Request): [description]

        Returns:
            web.Response: [description]
        """
        return web.Response(text="Your data: {}".format(request.match_info["data"]))


@dataclass
class ExampleHTTP(ComponentConfig):
    """Example HTTP device."""

    foo: bool = False
    bar: Optional[int] = 10

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=ExampleHTTPDevice(foo=self.foo, bar=self.bar),
            adapters=[ExampleHTTPAdapter()],
        )



from dataclasses import dataclass
from typing import Hashable, Iterator, Mapping, NewType, Optional

from apischema import deserializer, serializer
from immutables import Map

#: An identifier which specifies the component
ComponentID = NewType("ComponentID", str)
#: An identifier which specifies the input/output port of a component
PortID = NewType("PortID", str)
#: A mapping of PortID to component input/output property
State = NewType("State", Mapping[PortID, Hashable])
#: A mapping of the difference between State mappings
Changes = NewType("Changes", Map[PortID, Hashable])
#: Simulation Time in nanoseconds
SimTime = NewType("SimTime", int)


@dataclass(frozen=True)
class ComponentPort:
    """An immutable dataclass for custom (de)serialization of component - port pairs."""

    component: ComponentID
    port: PortID

    def __repr__(self) -> str:
        """A string representation of the object of format component:port.

        Returns:
            str: A string representation of the object of format component:port.
        """
        return ":".join((self.component, self.port))

    @serializer
    def serialize(self) -> str:
        """An apischema serialization method which returns a string of component:port.

        Returns:
            str: The serialized ComponentPort, in format component:port.
        """
        return str(self)

    @deserializer
    @staticmethod
    def deserialize(data: str) -> "ComponentPort":
        """An apischema deserialization method which builds a from a string of
        component:port.

        Returns:
            ComponentPort: The deserialized ComponentPort.
        """
        component, port = data.split(":")
        return ComponentPort(ComponentID(component), PortID(port))

    def __iter__(self) -> Iterator:
        """An iterator which returns (component, port).

        Returns:
            Iterator: An iterator containing (component, port).
        """
        return (x for x in (self.component, self.port))


@dataclass(frozen=True)
class Input:
    """An immutable data container for Component inputs.

    Args:
        target: The Component which is to handle the Input.
        time: The simulation time at which the Input was produced and is to be handled.
        changes: The changes to the component inputs.
    """

    target: ComponentID
    time: SimTime
    changes: Changes


@dataclass(frozen=True)
class Output:
    """An immutable data container for Component outputs.

    Args:
        source: The Component which produced the Output.
        time: The simulation time at which the Output was produced and is to be handled.
        changes: The changes to the component outputs.
        call_in: The duration after which the component requests to be awoken.
    """

    source: ComponentID
    time: SimTime
    changes: Changes
    call_at: Optional[SimTime]


@dataclass(frozen=True)
class Interrupt:
    """An immutable data container for scheduling Component interrupts.

    Args:
        component: The Component which is requesting an interrupt.
    """

    source: ComponentID



from dataclasses import dataclass

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation

from .pneumatic import PneumaticAdapter, PneumaticDevice


@dataclass
class Pneumatic(ComponentConfig):
    """Pneumatic simulation with EPICS IOC adapter."""

    initial_speed: float = 2.5
    initial_state: bool = False
    db_file: str = "tickit/devices/pneumatic/db_files/filter1.db"
    ioc_name: str = "PNEUMATIC"

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=PneumaticDevice(
                initial_speed=self.initial_speed, initial_state=self.initial_state
            ),
            adapters=[PneumaticAdapter(db_file=self.db_file, ioc_name=self.ioc_name)],
        )



from abc import ABC, abstractmethod
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Generic,
    Optional,
    Tuple,
    TypeVar,
)

from typing_extensions import Protocol

from tickit.core.device import Device
from tickit.utils.configuration.configurable import as_tagged_union

#: Message type
T = TypeVar("T")


# https://github.com/python/mypy/issues/708#issuecomment-647124281
class RaiseInterrupt(Protocol):
    """A raise_interrupt function that should be passed to `Adapter`."""

    async def __call__(self) -> None:
        """The actual call signature."""
        pass


@as_tagged_union
class Adapter:
    """An interface for types which implement device adapters."""

    device: Device
    raise_interrupt: RaiseInterrupt

    def __getattr__(self, name: str) -> Any:
        """Improve error message for getting attributes before `run_forever`."""
        if name in ("device", "raise_interrupt"):
            raise RuntimeError(
                "Can't get self.device or self.raise_interrupt before run_forever()"
            )
        return super().__getattribute__(name)

    async def run_forever(
        self, device: Device, raise_interrupt: RaiseInterrupt
    ) -> None:
        """An asynchronous method allowing indefinite running of core adapter logic.

        An asynchronous method allowing for indefinite running of core adapter logic
        (typically the hosting of a protocol server and the interpretation of commands
        which are supplied via it).
        """
        self.device = device
        self.raise_interrupt = raise_interrupt

    def after_update(self):
        """A method which is called immediately after the device updates."""


@as_tagged_union
class Interpreter(ABC, Generic[T]):
    """An interface for types which handle messages recieved by an adapter."""

    @abstractmethod
    async def handle(
        self, adapter: Adapter, message: T
    ) -> Tuple[AsyncIterable[T], bool]:
        """An asynchronous method which handles messages recieved by an adapter.

        An asynchronous method which handles messages recieved by an adapter, replies
        are sent as an asynchronous iterable to support setting of continious readback,
        stand alone replies should be wrapped in an asynchronous iterable of length one.

        Args:
            adapter (Adapter): The adapter which is delegating message handling.
            message (T): The message recieved by the adapter.

        Returns:
            Tuple[AsyncIterable[T], bool]: A tuple containing both an asynchronous
                iterable of reply messages and an interrupt flag.
        """


@as_tagged_union
class Server(Generic[T]):
    """An interface for types which implement an external messaging protocol."""

    async def run_forever(
        self,
        on_connect: Callable[[], AsyncIterable[Optional[T]]],
        handler: Callable[[T], Awaitable[AsyncIterable[Optional[T]]]],
    ) -> None:
        """An asynchronous method allowing indefinite running of core server logic.

        Args:
            on_connect (Callable[[], AsyncIterable[Optional[T]]]): An asynchronous
                iterable of messages to be sent once a client connects.
            handler (Callable[[T], Awaitable[AsyncIterable[Optional[T]]]]): An
                asynchronous method used to handle recieved messages, returning an
                asynchronous iterable of replies.
        """



from typing import AnyStr, AsyncIterable, List, Tuple

import pytest
from mock import ANY, AsyncMock, call, patch

from tickit.adapters.interpreters.utils import wrap_as_async_iterable
from tickit.adapters.interpreters.wrappers import SplittingInterpreter
from tickit.core.adapter import Adapter


async def dummy_response(msg: AnyStr) -> Tuple[AsyncIterable[AnyStr], bool]:
    return wrap_as_async_iterable(msg), True


class DummySplittingInterpreter(SplittingInterpreter):
    """SplittingInterpreter with dummy implementation of abstract method."""

    async def _handle_individual_messages(
        self, adapter: Adapter, individual_messages: List[AnyStr]
    ) -> List[Tuple[AsyncIterable[AnyStr], bool]]:
        return [await dummy_response(msg) for msg in individual_messages]


async def _test_sub_messages(
    splitting_interpreter: DummySplittingInterpreter,
    mock_handle_individual_messages: AsyncMock,
    mock_get_response: AsyncMock,
    test_message: AnyStr,
    expected_sub_messages: List[AnyStr],
):
    mock_get_response.return_value = await dummy_response(test_message)
    await splitting_interpreter.handle(AsyncMock(), test_message)
    mock_handle_individual_messages.assert_called_once_with(ANY, expected_sub_messages)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "test_message, message_delimiter, expected_sub_messages",
    [
        ("test message", " ", ["test", "message"]),
        (b"foo/bar", b"/", [b"foo", b"bar"]),
        ("single message", "/", ["single message"]),
        ("just1the2words", r"[1-4]", ["just", "the", "words"]),
        ("#1J=1 #2 P", r"(#[1-8])|\s", ["", "#1", "J=1", "", "#2", "", "P"]),
    ],
)
@patch.object(DummySplittingInterpreter, "_collect_responses")
@patch.object(DummySplittingInterpreter, "_handle_individual_messages")
async def test_handle_passes_on_correct_sub_messages(
    mock_handle_individual_messages: AsyncMock,
    mock_get_response: AsyncMock,
    test_message: AnyStr,
    message_delimiter: AnyStr,
    expected_sub_messages: List[AnyStr],
):
    splitting_interpreter = DummySplittingInterpreter(
        AsyncMock(),
        message_delimiter,
    )

    await _test_sub_messages(
        splitting_interpreter,
        mock_handle_individual_messages,
        mock_get_response,
        test_message,
        expected_sub_messages,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("sub_messages", [["one", "two", "three"], ["test", "message"]])
async def test_handle_individual_messages_makes_correct_handle_calls(sub_messages):
    mock_interpreter = AsyncMock()
    splitting_interpreter = SplittingInterpreter(mock_interpreter, " ")
    await splitting_interpreter._handle_individual_messages(AsyncMock(), sub_messages)
    mock_handle_calls = mock_interpreter.handle.mock_calls
    assert mock_handle_calls == [call(ANY, msg) for msg in sub_messages]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    [
        "individual_messages",
        "individual_interrupts",
        "expected_combined_interrupt",
    ],
    [
        (["one", "two"], [False, False], False),
        (["three", "four"], [False, True], True),
        (["five", "six"], [True, True], True),
    ],
)
@patch(
    "tickit.adapters.interpreters.wrappers."
    "splitting_interpreter.wrap_messages_as_async_iterable"
)
async def test_individual_results_combined_correctly(
    mock_wrap_message: AsyncMock,
    individual_messages: List[AnyStr],
    individual_interrupts: List[bool],
    expected_combined_interrupt: bool,
):
    splitting_interpreter = DummySplittingInterpreter(AsyncMock(), " ")
    individual_results = [
        (wrap_as_async_iterable(msg), interrupt)
        for msg, interrupt in zip(individual_messages, individual_interrupts)
    ]
    (
        _,
        combined_interrupt,
    ) = await splitting_interpreter._collect_responses(individual_results)
    mock_wrap_message.assert_called_once_with(individual_messages)
    assert combined_interrupt == expected_combined_interrupt


@pytest.mark.asyncio
@patch(
    "tickit.adapters.interpreters.wrappers."
    "splitting_interpreter.wrap_messages_as_async_iterable"
)
async def test_multi_resp(
    mock_wrap_message: AsyncMock,
):
    async def multi_resp(msgs):
        for msg in msgs:
            yield msg

    splitting_interpreter = DummySplittingInterpreter(AsyncMock(), " ")
    individual_results = [(multi_resp(["resp1", "resp2"]), True)]
    await splitting_interpreter._collect_responses(individual_results)
    mock_wrap_message.assert_called_once_with(["resp1", "resp2"])


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ["message_delimiter", "response_delimiter"], [(" ", "/"), ("/", " ")]
)
@patch.object(DummySplittingInterpreter, "_collect_responses")
@patch.object(DummySplittingInterpreter, "_handle_individual_messages")
async def test_handles_empty_messages_correctly(
    mock_handle_individual_messages: AsyncMock,
    mock_collect: AsyncMock,
    message_delimiter: str,
    response_delimiter: str,
):
    splitting_interpreter = DummySplittingInterpreter(AsyncMock(), message_delimiter)
    mock_collect.return_value = "", False

    await splitting_interpreter.handle(AsyncMock(), "")

    mock_handle_individual_messages.assert_called_once_with(ANY, [""])


@pytest.mark.parametrize(
    ["message", "message_delimiter"], [(" ", " "), (" ", r"\s"), ("delim", "delim")]
)
@patch.object(DummySplittingInterpreter, "_collect_responses")
@patch.object(DummySplittingInterpreter, "_handle_individual_messages")
@pytest.mark.asyncio
async def test_delimiter_only_message_results_in_empty_messages_handled(
    mock_handle_individual_messages: AsyncMock,
    mock_collect: AsyncMock,
    message: str,
    message_delimiter: str,
):
    splitting_interpreter = DummySplittingInterpreter(
        AsyncMock(),
        message_delimiter,
    )
    mock_collect.return_value = "", False

    await splitting_interpreter.handle(AsyncMock(), message)

    mock_handle_individual_messages.assert_called_once_with(ANY, ["", ""])



import logging
from dataclasses import dataclass
from random import randint

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.compat.typing_compat import TypedDict

LOGGER = logging.getLogger(__name__)


class TrampolineDevice(Device):
    """A trivial toy device which requests a callback every update."""

    #: An empty typed mapping of device inputs
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: An empty typed mapping of device outputs
    Outputs: TypedDict = TypedDict("Outputs", {})

    def __init__(self, callback_period: int = int(1e9)) -> None:
        """A constructor of the sink which configures the device callback period.

        Args:
            callback_period (int): The simulation time callback period of the device
                (in nanoseconds). Defaults to int(1e9).
        """
        self.callback_period = SimTime(callback_period)

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which prints the inputs and requests a callback.

        The update method which prints the time of the update and the inputs then
        returns an empty output mapping and a request to be called back after the
        configured callback period.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which never contains any changes, and
                requests a callback after the configured callback period.
        """
        LOGGER.debug("Boing! ({}, {})".format(time, inputs))
        return DeviceUpdate(
            TrampolineDevice.Outputs(), SimTime(time + self.callback_period)
        )


class RandomTrampolineDevice(Device):
    """A trivial toy device which produced a random output and requests a callback."""

    #: An empty typed mapping of device inputs
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: A typed mapping containing the 'output' output value
    Outputs: TypedDict = TypedDict("Outputs", {"output": int})

    def __init__(self, callback_period: int = int(1e9)) -> None:
        """A constructor of the sink which configures the device callback period.

        Args:
            callback_period (int): The simulation time callback period of the device
                (in nanoseconds). Defaults to int(1e9).
        """
        self.callback_period = SimTime(callback_period)

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which produces a random output and requests a callback.

        The update method which prints the time of the update, the inputs and the
        output which will be produced then returns the random output value and a
        request to be called back after the configured callback period.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the random output,
                and requests a callback after the configured callback period.
        """
        output = randint(0, 255)
        LOGGER.debug(
            "Boing! (delta: {}, inputs: {}, output: {})".format(time, inputs, output)
        )
        return DeviceUpdate(
            RandomTrampolineDevice.Outputs(output=output),
            SimTime(time + self.callback_period),
        )


@dataclass
class RandomTrampoline(ComponentConfig):
    """Random thing that goes boing."""

    callback_period: int = int(1e9)

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=RandomTrampolineDevice(callback_period=self.callback_period),
        )



from typing import Iterable

import pytest
from mock import Mock, patch

from tickit.core.typedefs import SimTime
from tickit.devices.source import SourceDevice


@pytest.fixture
def source() -> SourceDevice:
    return SourceDevice(value=42)


@pytest.fixture
def mock_logging() -> Iterable[Mock]:
    with patch("tickit.devices.source.logger", autospec=True) as mock:
        yield mock


def test_source_constructor_method(source: SourceDevice):
    assert source.value == 42


def test_source_update_method(source: SourceDevice):
    device_update = source.update(SimTime(0), inputs={})
    assert device_update.outputs["value"] == 42
    assert device_update.call_at is None



from typing import Dict, cast

import pytest
from mock import MagicMock, Mock, create_autospec, mock_open, patch

from tickit.adapters.epicsadapter import EpicsAdapter, InputRecord
from tickit.core.adapter import Adapter, Interpreter
from tickit.core.device import Device


@pytest.fixture
def MockDevice() -> Mock:
    return create_autospec(Device, instance=False)


@pytest.fixture
def MockInterpreter() -> Mock:
    return MagicMock(Interpreter, instance=True)


@pytest.fixture
def epics_adapter() -> EpicsAdapter:
    return EpicsAdapter("db_file", "ioc_name")  # type: ignore


@pytest.fixture
def input_record() -> InputRecord:
    def setter():
        return None

    def getter():
        return False

    return InputRecord("input", Mock(setter), Mock(getter))


def test_epics_adapter_is_adapter():
    assert issubclass(EpicsAdapter, Adapter)


def test_epics_adapter_constuctor(epics_adapter: EpicsAdapter):
    pass


def test_epics_adapter_interrupt_records_empty_on_construct():
    with pytest.raises(AttributeError):
        EpicsAdapter.interrupt_records


def test_link_input_on_interrupt(
    epics_adapter: EpicsAdapter, input_record: InputRecord
):
    def getter():
        return False

    getter = Mock(getter)
    epics_adapter.interrupt_records = {}
    epics_adapter.link_input_on_interrupt(input_record, getter)
    assert epics_adapter.interrupt_records[input_record] == getter


def test_epics_adapter_after_update_method(
    epics_adapter: EpicsAdapter, input_record: InputRecord
):
    def getter():
        return False

    getter = Mock(getter)
    epics_adapter.interrupt_records = {}
    epics_adapter.interrupt_records[input_record] = getter

    epics_adapter.after_update()

    cast(Mock, input_record.set).assert_called()


def test_epics_adapter_on_db_load_method(epics_adapter: EpicsAdapter):
    with pytest.raises(NotImplementedError):
        epics_adapter.on_db_load()


def record_db_file_contents() -> Dict[str, bytes]:
    return {
        "data": b"""record(ao, "$(device):GAIN") {
        field(DTYP, "Hy8001")
        field(OMSL, "supervisory")
        field(OUT, "#C1 S0 @")
        field(DESC, "Gain value")
        field(EGU, "A")
        }""",
        "expected": b"""record(ao, "$(device):GAIN") {
        field(OMSL, "supervisory")
        field(OUT, "#C1 S0 @")
        field(DESC, "Gain value")
        field(EGU, "A")
        }""",
    }


def filter1_db_file_contents() -> Dict[str, bytes]:
    return {
        "data": b"""record(bo, "$(device):FILTER") {
        field(DTYP, "$(DTYP)")
        field(SCAN, "Passive")
        field(ZNAM, "Out")
        field(ONAM, "In")
        field(VAL, "0")
        field(OMSL, "closed_loop")
        }""",
        "expected": b"""record(bo, "$(device):FILTER") {
        field(SCAN, "Passive")
        field(ZNAM, "Out")
        field(ONAM, "In")
        field(VAL, "0")
        field(OMSL, "closed_loop")
        }""",
    }


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(record_db_file_contents(), id="record.db file contents"),
        pytest.param(filter1_db_file_contents(), id="filter.db file contents"),
    ],
)
def test_epics_adapter_load_records_without_DTYP_fields_method(
    epics_adapter: EpicsAdapter,
    test_params: Dict[str, bytes],
):
    data = test_params["data"]
    with patch("builtins.open", mock_open(read_data=data)):
        with patch("os.unlink") as mock_unlink:
            epics_adapter.load_records_without_DTYP_fields()
            unlink_args = mock_unlink.call_args.args

    out_filename = unlink_args[0]
    with open(out_filename, "rb") as file:
        written_data = file.read()

    assert str(written_data).strip() == str(test_params["expected"]).strip()



from collections import defaultdict, deque
from typing import DefaultDict, Dict, Iterable, Mapping, Optional, Set, Union, overload

from tickit.core.components.component import ComponentConfig
from tickit.core.typedefs import ComponentID, ComponentPort, PortID
from tickit.utils.compat.functools_compat import cached_property

#: A mapping of component output ports to component input ports with defaults
Default_Wiring_Struct = DefaultDict[
    ComponentID, DefaultDict[PortID, Set[ComponentPort]]
]
#: A mapping of component output ports to component input ports
Wiring_Struct = Dict[ComponentID, Dict[PortID, Set[ComponentPort]]]
#: A mapping of component input ports to component output ports with defaults
Default_InverseWiring_Struct = DefaultDict[
    ComponentID, DefaultDict[PortID, ComponentPort]
]
#: A mapping of component input ports to component output ports
Inverse_Wiring_Struct = Dict[ComponentID, Dict[PortID, ComponentPort]]


class Wiring(Default_Wiring_Struct):
    """A mapping of component output ports to component input ports with defaults.

    A mapping of component output ports to component input ports, used to represent the
    connections between components in a system. Defaults are generated in case of
    missing mapping keys; for an unknown component, an empty output port mapping is
    created, whilst for an unknown output port an empty input port set is created.
    """

    def __init__(self, wiring: Optional[Wiring_Struct] = None) -> None:
        """A constructor which adds defaults to a wiring struct if provided.

        A constructor which adds defaults to a wiring struct if provided, otherwise an
        empty default dict is created.

        Args:
            wiring (Optional[Wiring_Struct]): An optional wiring struct. Defaults to
                None.
        """
        _wiring = (
            {dev: DefaultDict(set, io) for dev, io in wiring.items()}
            if wiring
            else dict()
        )
        return super().__init__(lambda: DefaultDict(set, dict()), _wiring)

    @classmethod
    def from_inverse_wiring(cls, inverse_wiring: "InverseWiring") -> "Wiring":
        """Un-inverts an inverse wiring object to produce a Wiring instance.

        Args:
            inverse_wiring (InverseWiring): A mapping of component input ports to
                component output ports.

        Returns:
            Wiring: A mapping of component output ports to component input ports.
        """
        wiring: Wiring = cls()
        for in_dev, in_ios in inverse_wiring.items():
            wiring[in_dev]
            for in_io, (out_dev, out_io) in in_ios.items():
                wiring[out_dev][out_io].add(ComponentPort(in_dev, in_io))
        return wiring


class InverseWiring(Default_InverseWiring_Struct):
    """A mapping of component input ports to component output ports.

    A mapping of component input ports to component output ports, used to represent the
    connections between components in a system. When either an unknown input component
    or an unknown input component port is requested a KeyError is raised.
    """

    def __init__(self, wiring: Optional[Inverse_Wiring_Struct] = None) -> None:
        """A constructor which adds defaults to an inverse wiring struct if provided.

        A constructor which adds defaults to an inverse wiring struct if provided,
        otherwise an empty default dict is created.

        Args:
            wiring (Optional[Inverse_Wiring_Struct]): An optional inverse wiring struct.
                Defaults to None.
        """
        _wiring = (
            {dev: DefaultDict(None, io) for dev, io in wiring.items()}
            if wiring
            else dict()
        )
        return super().__init__(lambda: DefaultDict(None, dict()), _wiring)

    @classmethod
    def from_wiring(cls, wiring: Wiring) -> "InverseWiring":
        """Inverts a Wiring object to create an InverseWiring instance.

        Args:
            wiring (Wiring): A mapping of component output ports to component input
                ports.

        Returns:
            InverseWiring: A mapping of component input ports to component output ports.
        """
        inverse_wiring: InverseWiring = cls()
        for out_dev, out_ids in wiring.items():
            inverse_wiring[out_dev]
            for out_io, ports in out_ids.items():
                for in_dev, in_io in ports:
                    inverse_wiring[in_dev][in_io] = ComponentPort(out_dev, out_io)
        return inverse_wiring

    @classmethod
    def from_component_configs(
        cls, configs: Iterable[ComponentConfig]
    ) -> "InverseWiring":
        """Creates an InverseWiring instance from an iterable of ComponentConfigs.

        Args:
            configs (Iterable[ComponentConfig]): An iterable of component config data
                containers.

        Returns:
            InverseWiring: A mapping of component input ports to component output ports.
        """
        return cls({config.name: config.inputs for config in configs})


class EventRouter:
    """A utility class responsible for routing changes between components."""

    _wiring: Wiring

    @overload
    def __init__(self, wiring: Wiring) -> None:
        """Placeholder."""  # See: https://github.com/PyCQA/pydocstyle/issues/525
        pass

    @overload
    def __init__(self, wiring: InverseWiring) -> None:
        """Placeholder."""  # See: https://github.com/PyCQA/pydocstyle/issues/525
        pass

    def __init__(self, wiring: Union[Wiring, InverseWiring]) -> None:
        """Construct an EventRouter.

        An EventRouter stores possible inverse wiring for use in utilities.

        Args:
            wiring (Union[Wiring, InverseWiring]): A wiring or inverse wiring object
                representing the connections between components in the system.
        """
        if isinstance(wiring, Wiring):
            self._wiring = wiring
        elif isinstance(wiring, InverseWiring):
            self._wiring = Wiring.from_inverse_wiring(wiring)

    @cached_property
    def wiring(self) -> Wiring:
        """The cached wiring used by the event router.

        Returns:
            Wiring: The wiring used by the event router.
        """
        return self._wiring

    @cached_property
    def components(self) -> Set[ComponentID]:
        """A cached set of all components in the wiring.

        Returns:
            Set[ComponentID]: A set of all components in the wiring.
        """
        return set.union(self.input_components, self.output_components)

    @cached_property
    def output_components(self) -> Set[ComponentID]:
        """A cached set of components which provide outputs.

        Returns:
            Set[ComponentID]: A set of components which provide outputs.
        """
        return {component for component, port in self.wiring.items() if port}

    @cached_property
    def input_components(self) -> Set[ComponentID]:
        """A cached set of components which recieve inputs.

        Returns:
            Set[ComponentID]: A set of components which recieve inputs.
        """
        return set(
            dev
            for out in self.wiring.values()
            for port in out.values()
            for dev, _ in port
        )

    @cached_property
    def component_tree(self) -> Dict[ComponentID, Set[ComponentID]]:
        """A cached mapping of first order component dependants.

        A cached property which returns a mapping of components to the set of
        components which are wired to any of its outputs.

        Returns:
            Dict[ComponentID, Set[ComponentID]]:
                A mapping of components to the set of components which are wired to any
                of its outputs.
        """
        return {
            dev: set(dev for port in out.values() for dev, _ in port)
            for dev, out in self.wiring.items()
        }

    @cached_property
    def inverse_component_tree(self) -> Dict[ComponentID, Set[ComponentID]]:
        """A cached mapping of first order component dependancies.

        A cached property which returns a mapping of components to the set of
        components which are wired to any of its inputs.

        Returns:
            Dict[ComponentID, Set[ComponentID]]:
                A mapping of components to the set of components which are wired to any
                of its inputs.
        """
        inverse_tree: Dict[ComponentID, Set[ComponentID]] = {
            dev: set() for dev in self.components
        }
        for dev, deps in self.component_tree.items():
            for dep in deps:
                inverse_tree[dep].add(dev)
        return inverse_tree

    def dependants(self, root: ComponentID) -> Set[ComponentID]:
        """Finds the set of all components which are recursively dependant on the root
        component.

        Args:
            root (ComponentID): The root component.

        Returns:
            Set[ComponentID]:
                A set of all components which are recursively dependant on the root
                component.
        """
        dependants = set()
        to_crawl = deque([root])
        while to_crawl:
            dev = to_crawl.popleft()
            if dev not in dependants:
                dependants.add(dev)
                if dev in self.component_tree.keys():
                    to_crawl.extend(self.component_tree[dev] - dependants)
        return dependants

    def route(
        self, source: ComponentID, changes: Mapping[PortID, object]
    ) -> DefaultDict[ComponentID, Dict[PortID, object]]:
        """Generates a mapping of input changes resulting from output changes.

        A method which generates a mapping of input changes which result from the
        propagation of output changes according to the wiring.

        Args:
            source (ComponentID): The source component of the output changes.
            changes (Mapping[PortID, Hashable]): A mapping of changes to the outputs of
                the source component.

        Returns:
            DefaultDict[ComponentID, Dict[PortID, Hashable]]:
                A mapping of the input changes which result from the propagation of
                output changes according to the wiring.
        """
        routed: DefaultDict[ComponentID, Dict[PortID, object]] = defaultdict(dict)
        for out_id, out_val in changes.items():
            for in_dev, in_id in self.wiring[source][out_id]:
                routed[in_dev][in_id] = out_val
        return routed



from collections import defaultdict
from typing import List

import pytest
from immutables import Map
from mock import AsyncMock, MagicMock

from tickit.core.management.event_router import Inverse_Wiring_Struct, InverseWiring
from tickit.core.management.ticker import Ticker
from tickit.core.typedefs import (
    Changes,
    ComponentID,
    ComponentPort,
    Input,
    Output,
    PortID,
    SimTime,
)


@pytest.fixture
def inverse_wiring_struct() -> Inverse_Wiring_Struct:
    return {
        ComponentID("Mid1"): {
            PortID("Mid1<1"): ComponentPort(ComponentID("Out1"), PortID("Out1>1")),
            PortID("Mid1<2"): ComponentPort(ComponentID("Ext1"), PortID("Ext1>1")),
        },
        ComponentID("In1"): {
            PortID("In1<1"): ComponentPort(ComponentID("Mid1"), PortID("Mid1>1")),
            PortID("In1<2"): ComponentPort(ComponentID("Ext1"), PortID("Ext1>1")),
        },
    }


@pytest.fixture
def inverse_wiring(inverse_wiring_struct: Inverse_Wiring_Struct) -> InverseWiring:
    return InverseWiring(inverse_wiring_struct)


@pytest.fixture
def ticker(inverse_wiring: InverseWiring) -> Ticker:
    return Ticker(inverse_wiring, AsyncMock())


def test_ticker_components_returns_components(ticker: Ticker):
    assert {
        ComponentID("In1"),
        ComponentID("Mid1"),
        ComponentID("Out1"),
        ComponentID("Ext1"),
    } == ticker.components


@pytest.mark.asyncio
async def test_ticker_schedule_possible_updates_schedules_only_possible(ticker: Ticker):
    ticker.time = SimTime(42)
    ticker.inputs = defaultdict(dict)
    ticker.roots = {ComponentID("Out1"), ComponentID("Mid1")}
    ticker.to_update = {ComponentID("Out1"): None, ComponentID("Mid1"): None}
    ticker.update_component = AsyncMock()
    await ticker.schedule_possible_updates()
    ticker.update_component.assert_called_once_with(
        Input(ComponentID("Out1"), SimTime(42), Changes(Map()))
    )


@pytest.mark.asyncio
async def test_ticker_schedule_possible_updates_passes_inputs(ticker: Ticker):
    ticker.time = SimTime(42)
    ticker.inputs = defaultdict(
        dict, {ComponentID("Out1"): {PortID("TestChange"): 3.14}}
    )
    ticker.roots = set()
    ticker.to_update = {ComponentID("Out1"): None}
    ticker.update_component = AsyncMock()
    await ticker.schedule_possible_updates()
    ticker.update_component.assert_called_once_with(
        Input(
            ComponentID("Out1"), SimTime(42), Changes(Map({PortID("TestChange"): 3.14}))
        )
    )


@pytest.mark.asyncio
async def test_ticker_propagate_raises_unexpected_output(ticker: Ticker):
    ticker.time = SimTime(42)
    ticker.to_update = {ComponentID("Out1"): None}
    with pytest.raises(AssertionError):
        await ticker.propagate(
            Output(ComponentID("Mid1"), SimTime(42), Changes(Map()), None)
        )


@pytest.mark.asyncio
async def test_ticker_propagate_raises_unexpected_time(ticker: Ticker):
    ticker.time = SimTime(42)
    ticker.to_update = {ComponentID("Out1"): None}
    with pytest.raises(AssertionError):
        await ticker.propagate(
            Output(ComponentID("Out1"), SimTime(64), Changes(Map()), None)
        )


@pytest.mark.asyncio
async def test_ticker_propagate_schedules_next(ticker: Ticker):
    ticker.time = SimTime(42)
    ticker.inputs = defaultdict(dict)
    ticker.roots = {ComponentID("Mid1")}
    ticker.to_update = {ComponentID("Out1"): None, ComponentID("Mid1"): None}
    ticker.update_component = AsyncMock()
    await ticker.propagate(
        Output(ComponentID("Out1"), SimTime(42), Changes(Map()), None)
    )
    ticker.update_component.assert_called_once_with(
        Input(ComponentID("Mid1"), SimTime(42), Changes(Map()))
    )


@pytest.mark.asyncio
async def test_ticker_propagate_sets_finished_once_complete(ticker: Ticker):
    ticker.time = SimTime(42)
    ticker.inputs = defaultdict(dict)
    ticker.roots = set()
    ticker.to_update = {ComponentID("Out1"): None}
    ticker.update_component = AsyncMock()
    await ticker.propagate(
        Output(ComponentID("Out1"), SimTime(42), Changes(Map()), None)
    )
    assert ticker.finished.is_set()


@pytest.mark.asyncio
async def test_ticker_start_tick_sets_time(ticker: Ticker):
    await ticker._start_tick(SimTime(42), set())
    assert SimTime(42) == ticker.time


@pytest.mark.asyncio
async def test_ticker_start_tick_sets_empty_inputs(ticker: Ticker):
    await ticker._start_tick(SimTime(42), set())
    assert defaultdict(dict, {}) == ticker.inputs


@pytest.mark.asyncio
async def test_ticker_start_tick_adds_root_component_dependents_to_to_update(
    ticker: Ticker,
):
    await ticker._start_tick(SimTime(42), {ComponentID("Mid1")})
    assert {ComponentID("Mid1"): None, ComponentID("In1"): None} == ticker.to_update


@pytest.mark.asyncio
async def test_ticker_call_starts_tick(ticker: Ticker):
    ticker.finished.set()
    ticker._start_tick = MagicMock(ticker._start_tick)  # type: ignore
    await ticker(SimTime(42), set())
    ticker._start_tick.assert_called_once()


@pytest.mark.asyncio
async def test_ticker_call_schedules_possible_updates(ticker: Ticker):
    ticker.finished.set()
    ticker.schedule_possible_updates = MagicMock(  # type: ignore
        ticker.schedule_possible_updates
    )
    await ticker(SimTime(42), set())
    ticker.schedule_possible_updates.assert_called_once()


@pytest.mark.asyncio
async def test_ticker_does_tick(ticker: Ticker):
    updates: List[Input] = list()

    async def update_component(input: Input) -> None:
        updates.append(input)
        if input.target == ComponentID("Mid1"):
            await ticker.propagate(
                Output(
                    input.target, input.time, Changes(Map({PortID("Mid1>1"): 42})), None
                )
            )
        await ticker.propagate(Output(input.target, input.time, Changes(Map()), None))

    ticker.update_component = update_component
    await ticker(SimTime(42), {ComponentID("Ext1"), ComponentID("Mid1")})
    assert [
        Input(ComponentID("Ext1"), SimTime(42), Changes(Map())),
        Input(ComponentID("Mid1"), SimTime(42), Changes(Map())),
        Input(ComponentID("In1"), SimTime(42), Changes(Map({PortID("In1<1"): 42}))),
    ] == updates



import sys
import warnings

if sys.version_info >= (3, 8):
    from functools import cached_property
elif sys.version_info >= (3, 5):
    cached_property = property
    warnings.warn(
        "Property caching is not performed when running with python versions prior to"
        + "3.8. Performance may be degraded.",
        RuntimeWarning,
    )

__all__ = ["cached_property"]



import pytest
from mock import ANY, AsyncMock

from tickit.adapters.interpreters.wrappers import BeheadingInterpreter


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "test_message, header_size, expected_beheaded_message",
    [
        ("headermessage", 6, "message"),
        (b"@\xbf\x00\x00\x00\x00\x00\x04message\r", 8, b"message\r"),
    ],
)
async def test_beheading_interpreter_removes_expected_number_of_characters(
    test_message, header_size, expected_beheaded_message
):
    mock_interpreter = AsyncMock()
    beheading_interpreter = BeheadingInterpreter(mock_interpreter, header_size)
    await beheading_interpreter.handle(AsyncMock(), test_message)
    mock_interpreter.handle.assert_called_once_with(ANY, expected_beheaded_message)



import pytest

from tickit.devices.eiger.filewriter.filewriter_config import FileWriterConfig

# # # # # Eiger FileWriterConfig Tests # # # # #


@pytest.fixture
def filewriter_config() -> FileWriterConfig:
    return FileWriterConfig()


def test_eiger_filewriter_config_constructor():
    FileWriterConfig()


def test_eiger_filewriter_config_getitem(filewriter_config):
    assert "enabled" == filewriter_config["mode"]["value"]



import re
from dataclasses import dataclass
from typing import AnyStr, Callable, Generic, Optional, Sequence


@dataclass(frozen=True)
class RegexCommand(Generic[AnyStr]):
    """A decorator to register an adapter method as a regex parsed command.

    Args:
        regex (Union[bytes, str]): The regular expression pattern which must be
            matched in full, with groups used to extract command arguments.
        interrupt (bool): A flag indicating whether calling of the method should
            raise an adapter interrupt. Defaults to False.
        format (Optional[str]): The message decoding format to be used for string
            based interpretation. Defaults to None.

    Returns:
        Callable:
            A decorator which registers the adapter method as a message handler.
    """

    regex: AnyStr
    interrupt: bool = False
    format: Optional[str] = None

    def __call__(self, func: Callable) -> Callable:
        """A decorator which registers the adapter method as a message handler.

        Args:
            func (Callable): The adapter method to be registered as a command.

        Returns:
            Callable: The registered adapter method.
        """
        setattr(func, "__command__", self)
        return func

    def parse(self, data: bytes) -> Optional[Sequence[AnyStr]]:
        """Performs message decoding and regex matching to match and extract arguments.

        A method which performs message decoding accoridng to the command formatting
        string, checks for a full regular expression match and returns a sequence of
        function arguments if a match is found, otherwise the method returns None.

        Args:
            data (bytes): The message data to be parsed.

        Returns:
            Optional[Sequence[AnyStr]]:
                If a full match is found a sequence of function arguments is returned,
                otherwise the method returns None.
        """
        message = data.decode(self.format, "ignore").strip() if self.format else data
        if isinstance(message, type(self.regex)):
            match = re.fullmatch(self.regex, message)
            if match:
                return match.groups()
        return None



from dataclasses import is_dataclass
from typing import Optional

from tickit.core.device import DeviceUpdate, OutMap
from tickit.core.typedefs import SimTime


def test_device_update_is_dataclass():
    assert is_dataclass(DeviceUpdate)


def test_device_update_has_state():
    assert OutMap == DeviceUpdate.__annotations__["outputs"]


def test_device_update_has_call_in():
    assert Optional[SimTime] == DeviceUpdate.__annotations__["call_at"]



from dataclasses import dataclass
from pathlib import Path
from typing import List


@dataclass
class Image:
    """Dataclass to create a basic Image object."""

    index: int
    hash: str
    dtype: str
    data: bytes
    encoding: str

    @classmethod
    def create_dummy_image(cls, index: int) -> "Image":
        """Returns an Image object wrapping the dummy blob using the metadata provided.

        Args:
            index (int): The index of the Image in the current acquisition.

        Returns:
            Image: An Image object wrapping the dummy blob.
        """
        data = dummy_image_blob()
        hsh = str(hash(data))
        dtype = "uint16"
        encoding = "bs16-lz4<"
        return Image(index, hsh, dtype, data, encoding)


_DUMMY_IMAGE_BLOBS: List[bytes] = []


def dummy_image_blob() -> bytes:
    """Returns the current dummy data blob.

    Return the raw bytes of a compressed image
    taken from the stream of a real Eiger detector.

    Returns:
        A compressed image as a bytes object.
    """
    if not _DUMMY_IMAGE_BLOBS:
        with open(
            Path(Path(__file__).parent.parent, "resources", "frame_sample"),
            "rb",
        ) as frame_file:
            _DUMMY_IMAGE_BLOBS.append(frame_file.read())
    return _DUMMY_IMAGE_BLOBS[0]



from typing import Any, Callable, Iterable, Type

import pytest
from apischema.conversions.conversions import Conversion
from mock import Mock, create_autospec, patch
from mock.mock import mock_open

from tickit.core.components.component import ComponentConfig
from tickit.core.typedefs import ComponentID, ComponentPort, PortID
from tickit.utils.configuration.loading import importing_conversion, read_configs


@pytest.fixture
def mock_component_config_type() -> Mock:
    return create_autospec(ComponentConfig, instance=False)


@pytest.fixture
def patch_tagged_union_dict(mock_component_config_type) -> Iterable[Mock]:
    with patch.dict(
        "tickit.utils.configuration.loading.is_tagged_union",
        {mock_component_config_type: True},
    ) as mock:
        yield mock


def test_importing_conversion(
    mock_component_config_type: Type[ComponentConfig], patch_tagged_union_dict
):
    conversion = importing_conversion(mock_component_config_type)
    assert isinstance(conversion, Conversion)
    assert conversion.target == mock_component_config_type


def test_importing_conversion_when_is_not_tagged_union(
    mock_component_config_type: Type[ComponentConfig],
):
    conversion = importing_conversion(mock_component_config_type)
    assert conversion == []


class MockConfig(ComponentConfig):
    pass

    def __call__(self):
        return Mock()


@pytest.fixture
def patch_apischema_deserialize() -> Iterable[Mock]:
    with patch(
        "tickit.utils.configuration.loading.deserialize",
        autospec=True,
    ) as mock:
        mock.return_value = [
            MockConfig(
                name=ComponentID("foo"),
                inputs={PortID("42"): ComponentPort(ComponentID("bar"), PortID("24"))},
            )
        ]
        yield mock


def test_conversion(
    mock_component_config_type,
    patch_tagged_union_dict,
    patch_apischema_deserialize: Mock,
):
    conversion = importing_conversion(mock_component_config_type)
    converter: Callable[[Any], Any] = conversion.converter  # type: ignore

    _ = converter({"mock.Mock": 42})
    patch_apischema_deserialize.assert_called_once_with(
        Mock, 42, default_conversion=importing_conversion
    )


@pytest.fixture
def patch_yaml_library() -> Iterable[Mock]:
    with patch("tickit.utils.configuration.loading.yaml", autospec=True) as mock:
        yield mock


@pytest.fixture
def patch_builtins_open() -> Iterable[Mock]:
    blank_yaml = ""
    with patch(
        "tickit.utils.configuration.loading.open",
        new=mock_open(read_data=blank_yaml),
    ) as mock:
        yield mock


def test_read_configs(
    patch_builtins_open: Mock,
    patch_apischema_deserialize: Mock,
):
    configs = read_configs("foo/bar.borg")
    assert isinstance(configs[0], MockConfig)



import pytest

from tickit.devices.eiger.eiger import EigerDevice
from tickit.devices.eiger.eiger_status import State


@pytest.fixture
def eiger() -> EigerDevice:
    return EigerDevice()


def test_eiger_constructor():
    EigerDevice()


@pytest.mark.asyncio
async def test_eiger_initialize(eiger: EigerDevice):
    await eiger.initialize()

    assert State.IDLE.value == eiger.get_state()["value"]


@pytest.mark.asyncio
async def test_eiger_arm(eiger: EigerDevice):
    await eiger.arm()

    assert State.READY.value == eiger.get_state()["value"]


@pytest.mark.asyncio
async def test_eiger_disarm(eiger: EigerDevice):
    await eiger.disarm()

    assert State.IDLE.value == eiger.get_state()["value"]


@pytest.mark.asyncio
async def test_eiger_trigger_ints_and_ready(eiger: EigerDevice):
    eiger._set_state(State.READY)
    eiger.settings.trigger_mode = "ints"

    message = await eiger.trigger()

    assert State.ACQUIRE.value == eiger.get_state()["value"]
    assert "Aquiring Data from Eiger..." == message


@pytest.mark.asyncio
async def test_eiger_trigger_not_ints_and_ready(eiger: EigerDevice):
    eiger._set_state(State.READY)

    message = await eiger.trigger()

    assert State.READY.value == eiger.get_state()["value"]
    assert (
        f"Ignoring trigger, state={eiger.status.state},"
        f"trigger_mode={eiger.settings.trigger_mode}" == message
    )


@pytest.mark.asyncio
async def test_eiger_trigger_not_ints_and_not_ready(eiger: EigerDevice):
    eiger._set_state(State.IDLE)

    message = await eiger.trigger()

    assert State.READY.value != eiger.get_state()["value"]
    assert (
        f"Ignoring trigger, state={eiger.status.state},"
        f"trigger_mode={eiger.settings.trigger_mode}" == message
    )


@pytest.mark.asyncio
async def test_eiger_cancel(eiger: EigerDevice):
    await eiger.cancel()

    assert State.READY.value == eiger.get_state()["value"]


@pytest.mark.asyncio
async def test_eiger_abort(eiger: EigerDevice):
    await eiger.abort()

    assert State.IDLE.value == eiger.get_state()["value"]


def test_eiger_get_state(eiger: EigerDevice):
    assert State.NA.value == eiger.get_state()["value"]


def test_eiger_set_state(eiger: EigerDevice):
    eiger._set_state(State.IDLE)

    assert State.IDLE.value == eiger.get_state()["value"]



from dataclasses import dataclass

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.devices.eiger.eiger import EigerDevice
from tickit.devices.eiger.eiger_adapters import EigerRESTAdapter, EigerZMQAdapter


@dataclass
class Eiger(ComponentConfig):
    """Eiger simulation with HTTP adapter."""

    host: str = "0.0.0.0"
    port: int = 8081
    zmq_host: str = "127.0.0.1"
    zmq_port: int = 9999

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=EigerDevice(),
            adapters=[
                EigerRESTAdapter(host=self.host, port=self.port),
                EigerZMQAdapter(zmq_host=self.zmq_host, zmq_port=self.zmq_port),
            ],
        )



import asyncio
import logging
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Coroutine,
    DefaultDict,
    Dict,
    Hashable,
    Optional,
    Set,
    Union,
)

from immutables import Map

from tickit.core.management.event_router import EventRouter, InverseWiring, Wiring
from tickit.core.typedefs import Changes, ComponentID, Input, Output, PortID, SimTime

LOGGER = logging.getLogger(__name__)


class Ticker:
    """A utility class responsible for sequencing the update of components during a
    tick.

    A utility class responsible for sequencing the update of components during a tick
    by eagerly updating each component which has had all of its dependencies resolved.
    """

    def __init__(
        self,
        wiring: Union[Wiring, InverseWiring],
        update_component: Callable[[Input], Coroutine[Any, Any, None]],
    ) -> None:
        """A Ticker constructor which creates an event router and performs initial
        setup.

        Args:
            wiring (Union[Wiring, InverseWiring]): A wiring or inverse wiring object
                representing the connections between components in the system.
            update_component (Callable[[Input], Awaitable[None]]): A function or method
                which may be called to request a component performs and update, such
                updates should result in a subsequent call to the propagate method of
                the ticker.
        """
        self.event_router = EventRouter(wiring)
        self.update_component = update_component
        self.to_update: Dict[ComponentID, Optional[asyncio.Task]] = dict()
        self.finished: asyncio.Event = asyncio.Event()

    async def __call__(
        self, time: SimTime, update_components: Set[ComponentID]
    ) -> None:
        """Performs a tick which updates the provided components and their dependants.

        An asynchronous method which performs a tick by setting up the initial state of
        the system during the tick - including determining dependant components,
        scheduling updates which require no component resolutions to be performed,
        before blocking until the system is resolved by update propagation.

        Args:
            time (SimTime): The simulation time at which the tick occurs (in
                nanoseconds).
            update_components (Set[ComponentID]): A set of components which require
                update.
        """
        await self._start_tick(time, update_components)
        await self.schedule_possible_updates()
        await self.finished.wait()
        self.finished.clear()

    async def _start_tick(self, time: SimTime, update_components: Set[ComponentID]):
        """Sets up the ticker to perform a tick.

        An asynchronous method which sets up the ticker to perform a tick by updating
        time, reseting accumulators and finding the set of components which require
        update.

        Args:
            time (SimTime): The simulation time at which the tick occurs (in
                nanoseconds).
            update_components (Set[ComponentID]): A set of components which require
                update.
        """
        self.time = time
        self.roots = update_components
        LOGGER.debug("Doing tick @ {}".format(self.time))
        self.inputs: DefaultDict[ComponentID, Dict[PortID, Hashable]] = defaultdict(
            dict
        )
        self.to_update = {
            c: None
            for component in self.roots
            for c in self.event_router.dependants(component)
        }

    async def schedule_possible_updates(self) -> None:
        """Updates components with resolved dependencies.

        An asynchronous method which schedules updates for components with resolved
        dependencies, as determined by the intersection of the components first order
        dependencies and the set of componets which still require an update.
        """

        def required_dependencies(component) -> Set[ComponentID]:
            return self.event_router.inverse_component_tree[component].intersection(
                self.to_update
            )

        updating: Dict[ComponentID, asyncio.Task] = dict()
        skipping: Set[ComponentID] = set()
        for component, task in self.to_update.items():
            if task is not None or required_dependencies(component):
                continue
            if self.inputs[component] or component in self.roots:
                updating[component] = asyncio.create_task(
                    self.update_component(
                        Input(
                            component, self.time, Changes(Map(self.inputs[component]))
                        )
                    )
                )
            else:
                skipping.add(component)
        self.to_update.update(updating)
        for component in skipping:
            del self.to_update[component]

    async def propagate(self, output: Output) -> None:
        """Propagates the output of an updated component.

        An asynchronous message which propagates the output of an updated component by
        removing the component from the set of components requiring update, adding the
        routed inputs to the accumulator, and scheduling any possible updates. If no
        components require update the finsihed flag will be set.

        Args:
            output (Output): The output produced by the update of a component.
        """
        assert output.source in self.to_update.keys()
        assert output.time == self.time
        self.to_update.pop(output.source)

        changes = self.event_router.route(output.source, output.changes)
        for component, change in changes.items():
            self.inputs[component].update(change)

        await self.schedule_possible_updates()
        if not self.to_update:
            self.finished.set()

    @property
    def components(self) -> Set[ComponentID]:
        """The set of all components in the wiring.

        Returns:
            Set[ComponentID]: The set of all components in the wiring.
        """
        return self.event_router.components



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Hashable, Mapping, Optional, TypeVar

from tickit.core.typedefs import SimTime
from tickit.utils.configuration.configurable import as_tagged_union

#: A bound typevar for mappings of device inputs
InMap = TypeVar("InMap", bound=Mapping[str, Hashable])
#: A bound typevar for mappings of device outputs
OutMap = TypeVar("OutMap", bound=Mapping[str, Hashable])


@dataclass
class DeviceUpdate(Generic[OutMap]):
    """An immutable data container for Device outputs and callback request time.

    Args:
        outputs: A mapping of device output keys and current values.
        call_at: The simulation time at which the component requests to be awoken.
    """

    outputs: OutMap
    call_at: Optional[SimTime]


@as_tagged_union
class Device(ABC, Generic[InMap, OutMap]):
    """An interface for types which implement simulated devices."""

    @abstractmethod
    def update(self, time: SimTime, inputs: InMap) -> DeviceUpdate[OutMap]:
        """A method which implements device behaviour according to the time and its
        inputs.

        A method which implements the (typically physics based) changes which occur
        within the device in response to either the progression of time or the
        alteration of inputs to the device. The method returns the new observable state
        of the device and may optionally include a time in which the method should be
        called again.

        Args:
            time: The current simulation time (in nanoseconds).
            inputs: A mapping of device inputs and their values.
        """



import logging
from dataclasses import dataclass

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.compat.typing_compat import TypedDict

LOGGER = logging.getLogger(__name__)


@dataclass
class Counter(ComponentConfig):
    """Simulation of simple counting device."""

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(name=self.name, device=CounterDevice())


class CounterDevice(Device):
    """A simple device which increments a value."""

    #: An empty typed mapping of input values
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: A typed mapping containing the 'value' output value
    Outputs: TypedDict = TypedDict("Outputs", {"value": int})

    def __init__(self, initial_value: int = 0, callback_period: int = int(1e9)) -> None:
        """A constructor of the counter, which increments the input value.

        Args:
            initial_value (Any): The initial value of the counter.
            callback_period (int): The simulation time callback period of the device
                (in nanoseconds). Defaults to int(1e9).
        """
        self._value = initial_value
        self.callback_period = SimTime(callback_period)
        LOGGER.debug(f"Counter initialized with value => {self._value}")

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which produces the incremented value.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the incremented value, and
                requests a callback after callback_period.
        """
        self._value = self._value + 1
        LOGGER.debug("Counter incremented to {}".format(self._value))
        return DeviceUpdate(
            CounterDevice.Outputs(value=self._value),
            SimTime(time + self.callback_period),
        )



import logging
from abc import abstractmethod
from typing import Dict, Optional, Set, Tuple, Type, Union

from tickit.core.management.event_router import InverseWiring, Wiring
from tickit.core.management.ticker import Ticker
from tickit.core.state_interfaces import StateConsumer, StateProducer
from tickit.core.typedefs import ComponentID, Input, Interrupt, Output, SimTime
from tickit.utils.topic_naming import input_topic, output_topic

LOGGER = logging.getLogger(__name__)


class BaseScheduler:
    """A base scheduler class which implements logic common to all schedulers."""

    def __init__(
        self,
        wiring: Union[Wiring, InverseWiring],
        state_consumer: Type[StateConsumer],
        state_producer: Type[StateProducer],
    ) -> None:
        """A constructor which stores inputs and creates an empty wakeups mapping.

        Args:
            wiring (Union[Wiring, InverseWiring]): A wiring or inverse wiring object
                representing the connections between components in the system.
            state_consumer (Type[StateConsumer]): The state consumer class to be used
                by the component.
            state_producer (Type[StateProducer]): The state producer class to be used
                by the component.
        """
        self._wiring = wiring
        self._state_consumer_cls = state_consumer
        self._state_producer_cls = state_producer
        self.wakeups: Dict[ComponentID, SimTime] = dict()

    @abstractmethod
    async def schedule_interrupt(self, source: ComponentID) -> None:
        """An abstract asynchronous method which should schedule an interrupt
        immediately.

        Args:
            source (ComponentID): The component which should be updated.
        """
        raise NotImplementedError

    async def update_component(self, input: Input) -> None:
        """Sends a message with an input to the input topic of a component.

        Args:
            input (Input): The input to be included in the message sent to the
                component.
        """
        await self.state_producer.produce(input_topic(input.target), input)

    async def handle_message(self, message: Union[Interrupt, Output]) -> None:
        """A callback to handle interrupts or outputs produced by the state consumer.

        An asynchronous callback which handles interrupt and output messages produced by
        the state consumer; For Outputs, changes are propagated and wakeups scheduled
        if required, whilst handling of interrupts is deferred.

        Args:
            message (Union[Interrupt, Output]): An Interrupt or Output produced by the
                state consumer.
        """
        LOGGER.debug("Scheduler got {}".format(message))
        if isinstance(message, Output):
            await self.ticker.propagate(message)
            if message.call_at is not None:
                self.add_wakeup(message.source, message.call_at)
        if isinstance(message, Interrupt):
            await self.schedule_interrupt(message.source)

    async def setup(self) -> None:
        """Instantiates and configures the ticker and state interfaces.

        An asynchronous setup method which creates a ticker, a state consumer which is
        subscribed to the output topics of each component in the system, a state
        producer to produce component inputs.
        """
        self.ticker = Ticker(self._wiring, self.update_component)
        self.state_consumer: StateConsumer[
            Union[Interrupt, Output]
        ] = self._state_consumer_cls(self.handle_message)
        await self.state_consumer.subscribe(
            {output_topic(component) for component in self.ticker.components}
        )
        self.state_producer: StateProducer[Input] = self._state_producer_cls()

    def add_wakeup(self, component: ComponentID, when: SimTime) -> None:
        """Adds a wakeup to the mapping.

        Args:
            component (ComponentID): The component which should be updated.
            when (SimTime): The simulation time at which the update should occur.
        """
        LOGGER.debug("Scheduling {} for wakeup at {}".format(component, when))
        self.wakeups[component] = when

    def get_first_wakeups(self) -> Tuple[Set[ComponentID], Optional[SimTime]]:
        """Gets the components which are due for update first and the wakeup time.

        A method which returns a set of components which are due for update first and
        the simulation time at which the updates are scheduled. Or an empty set and
        None if no wakeups are scheduled.

        Returns:
            Tuple[Set[ComponentID], Optional[SimTime]]: A tuple containing the set of
                components which are scheduled for the first wakeup and the time at
                which the wakeup should occur. Or an empty set and None if no wakeups
                are scheduled.
        """
        if not self.wakeups:
            return set(), None
        first = min(self.wakeups.values())
        components = {
            component for component, when in self.wakeups.items() if when == first
        }
        return components, first



import logging
from typing import Awaitable, Callable, Dict, Optional, Set, Tuple, Type, Union

from immutables import Map

from tickit.core.management.event_router import InverseWiring, Wiring
from tickit.core.management.schedulers.base import BaseScheduler
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import (
    Changes,
    ComponentID,
    ComponentPort,
    Input,
    Output,
    PortID,
    SimTime,
)

LOGGER = logging.getLogger(__name__)


class SlaveScheduler(BaseScheduler):
    """A slave scheduler which orchestrates nested tickit simulations."""

    def __init__(
        self,
        wiring: Union[Wiring, InverseWiring],
        state_consumer: Type[StateConsumer],
        state_producer: Type[StateProducer],
        expose: Dict[PortID, ComponentPort],
        raise_interrupt: Callable[[], Awaitable[None]],
    ) -> None:
        """A slave scheduler constructor which adds wiring and saves values for
        reference.

        Args:
            wiring (Union[Wiring, InverseWiring]): A wiring or inverse wiring object
                representing the connections between components in the system.
            state_consumer (Type[StateConsumer]): The state consumer class to be used
                by the component.
            state_producer (Type[StateProducer]): The state producer class to be used
                by the component.
            expose (Dict[PortID, ComponentPort]): A mapping of slave scheduler
                outputs to internal component ports.
            raise_interrupt (Callable[[], Awaitable[None]]): A callback to request that
                the slave scheduler is updated immediately.
        """
        wiring = self.add_exposing_wiring(wiring, expose)
        super().__init__(wiring, state_consumer, state_producer)

        self.raise_interrupt = raise_interrupt
        self.interrupts: Set[ComponentID] = set()

    @staticmethod
    def add_exposing_wiring(
        wiring: Union[Wiring, InverseWiring],
        expose: Dict[PortID, ComponentPort],
    ) -> InverseWiring:
        """Adds wiring to expose slave scheduler outputs.

        Adds wiring to expose slave scheduler outputs, this is performed creating a
        mock "expose" component with inverse wiring set by expose.

        Args:
            wiring (Union[Wiring, InverseWiring]): A wiring or inverse wiring object
                representing the connections between components in the system.
            expose (Dict[PortID, ComponentPort]): A mapping of slave scheduler
                outputs to internal component ports.

        Returns:
            InverseWiring:
                An inverse wiring object representing the connections between
                components in the system and the "expose" component which acts as the
                slave scheduler output.
        """
        if isinstance(wiring, Wiring):
            wiring = InverseWiring.from_wiring(wiring)
        wiring[ComponentID("expose")].update(expose)
        return wiring

    async def update_component(self, input: Input) -> None:
        """Sends an input to the corresponding component. Mocks I/O for "external" or
        "expose".

        For real components the input is sent in a message to their input topic, for
        the mock component named "external", external inputs are injected, whilst for
        the mock component and named "expose" the input is stored for use as the
        scheduler output.

        Args:
            input (Input): The input message to be sent to the component.
        """
        if input.target == ComponentID("external"):
            await self.ticker.propagate(
                Output(ComponentID("external"), input.time, self.input_changes, None)
            )
        elif input.target == ComponentID("expose"):
            self.output_changes = input.changes
            await self.ticker.propagate(
                Output(ComponentID("expose"), input.time, Changes(Map()), None)
            )
        else:
            await super().update_component(input)

    async def on_tick(
        self, time: SimTime, changes: Changes
    ) -> Tuple[Changes, Optional[SimTime]]:
        """Routes inputs, performs a tick and returns output changes and a callback
        time.

        An asyhcnronous method which determines which components within the simulation
        require being woken up, sets the input changes for use by the "external" mock
        component, performs a tick, determines the period in which the slave scheduler
        should next be updated, and returns the changes collated by the "expose" mock
        component.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            changes (Changes): A mapping of changed component inputs and their new
                values.

        Returns:
            Tuple[Changes, Optional[SimTime]]:
                A tuple of a mapping of the changed exposed outputs and their new
                values and optionally a duration in simulation time after which the
                slave scheduler should be called again.
        """
        wakeup_components = {
            component for component, when in self.wakeups.items() if when <= time
        }
        root_components: Set[ComponentID] = {
            *self.interrupts,
            *wakeup_components,
            ComponentID("external"),
        }
        for component in wakeup_components:
            del self.wakeups[component]
        self.interrupts.clear()

        self.input_changes = changes
        self.output_changes = Changes(Map())
        await self.ticker(time, root_components)

        _, call_at = self.get_first_wakeups()
        return self.output_changes, call_at

    async def run_forever(self) -> None:
        """Delegates to setup which instantiates the ticker and state interfaces."""
        await self.setup()

    async def schedule_interrupt(self, source: ComponentID) -> None:
        """Schedules the interrupt of a component immediately.

        An asynchronous method which schedules an interrupt immediately by adding it to
        a set of queued interrupts and raising the interrupt to the master scheduler.

        Args:
            source (ComponentID): The source component of the interrupt.
        """
        LOGGER.debug("Adding {} to interrupts".format(source))
        self.interrupts.add(source)
        await self.raise_interrupt()



from typing import AnyStr, Optional, Tuple

import pytest

from tickit.adapters.interpreters.command import RegexCommand


@pytest.fixture
def regex_command(regex: AnyStr, interrupt: bool, format: Optional[str]):
    return RegexCommand(regex, interrupt, format)


def test_regex_command_registers_command():
    class TestAdapter:
        @RegexCommand("test")
        def test_method():
            pass

    assert isinstance(TestAdapter.test_method.__command__, RegexCommand)


@pytest.mark.parametrize(
    ["regex", "interrupt", "format", "message"],
    [
        (r"TestMessage", False, "utf-8", r"UnmatchedMessage".encode("utf-8")),
        (b"\\x01", False, None, b"\x02"),
    ],
)
def test_regex_command_parse_unmatched_returns_none(
    regex_command: RegexCommand, message: bytes
):
    assert regex_command.parse(message) is None


@pytest.mark.parametrize(
    ["regex", "interrupt", "format", "message", "expected"],
    [
        (r"TestMessage", False, "utf-8", r"TestMessage".encode("utf-8"), tuple()),
        (
            r"TestMessage(\d+)",
            False,
            "utf-8",
            r"TestMessage42".encode("utf-8"),
            ("42",),
        ),
        (b"\\x01", False, None, b"\x01", tuple()),
        (b"\\x01(.)", False, None, b"\x01\x02", (b"\x02",)),
    ],
)
def test_regex_command_parse_match_returns_args(
    regex_command: RegexCommand, message: bytes, expected: Tuple[object]
):
    assert expected == regex_command.parse(message)



import logging

from aiohttp import web
from apischema import serialize
from typing_extensions import TypedDict

from tickit.adapters.interpreters.endpoints.http_endpoint import HTTPEndpoint
from tickit.core.typedefs import SimTime
from tickit.devices.eiger.eiger_schema import Value
from tickit.devices.eiger.stream.stream_config import StreamConfig
from tickit.devices.eiger.stream.stream_status import StreamStatus

LOGGER = logging.getLogger(__name__)
STREAM_API = "stream/api/1.8.0"


class EigerStream:
    """Simulation of an Eiger stream."""

    stream_status: StreamStatus
    stream_config: StreamConfig
    stream_callback_period: SimTime

    #: An empty typed mapping of input values
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: A typed mapping containing the 'value' output value
    Outputs: TypedDict = TypedDict("Outputs", {})

    def __init__(self, callback_period: int = int(1e9)) -> None:
        """An Eiger Stream constructor."""
        self.stream_status = StreamStatus()
        self.stream_config = StreamConfig()
        self.stream_callback_period = SimTime(callback_period)


class EigerStreamAdapter:
    """An adapter for the Stream."""

    device: EigerStream

    @HTTPEndpoint.get(f"/{STREAM_API}" + "/status/{param}")
    async def get_stream_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting status values from the Stream.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]
        val = self.device.stream_status[param]["value"]
        meta = self.device.stream_status[param]["metadata"]

        data = serialize(
            Value(val, meta["value_type"].value, access_mode=meta["access_mode"])
        )

        return web.json_response(data)

    @HTTPEndpoint.get(f"/{STREAM_API}" + "/config/{param}")
    async def get_stream_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting config values from the Stream.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]
        val = self.device.stream_config[param]["value"]
        meta = self.device.stream_config[param]["metadata"]

        data = serialize(
            Value(val, meta["value_type"].value, access_mode=meta["access_mode"])
        )

        return web.json_response(data)
